#include "bsm_Mesh.h"
#include "bsfile.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

static void StoreTheMesh(void *userData,
      const char *name, int degree,
      int nv, const BSMvertex *mv, const int *mvhei,
      const point4d *vc,
      int nhe, const BSMhalfedge *mhe,
      int nfac, const BSMfacet *mfac, const int *mfhei,
      int spdimen, boolean rational, byte *mkv ) {
  int i;
  BSMesh *mesh = (BSMesh *) userData;
  assert (NULL != mesh);
  mesh->nv   = nv;
  mesh->nhe  = nhe;
  mesh->nfac = nfac;
  mesh->mv    = malloc ( nv*sizeof(BSMvertex) );
  mesh->mvhei = malloc ( nhe*sizeof(int) );
  mesh->mhe   = malloc ( nhe*sizeof(BSMhalfedge) );
  mesh->mfac  = malloc ( nfac*sizeof(BSMfacet) );
  mesh->mfhei = malloc ( nhe*sizeof(int) );
  switch(spdimen) {
    case 2:
      mesh->vc   = malloc (nv*sizeof(point2d) );
      break;
    case 3:
      mesh->vc = malloc (nv*sizeof(point3d) ); 
      break;
    default:
      exit(1);     
  }
  if ( !mesh->mv || !mesh->mvhei || !mesh->mhe || !mesh->mfac || !mesh->mfhei
|| !mesh->vc )
    exit ( 1 );
  memcpy ( mesh->mv, mv, nv*sizeof(BSMvertex) );
  memcpy ( mesh->mvhei, mvhei, nhe*sizeof(int) );
  memcpy ( mesh->mhe, mhe, nhe*sizeof(BSMhalfedge) );
  memcpy ( mesh->mfac, mfac, nfac*sizeof(BSMfacet) );
  memcpy ( mesh->mfhei, mfhei, nhe*sizeof(int) );
  if (2 == spdimen) { 
    for ( i = 0; i < nv; i++ ) {
      Point4to2d ( &vc[i], &(((point2d*)mesh->vc)[i]) );
    }
  } else if (3 == spdimen) {
    for ( i = 0; i < nv; i++ ) {
      Point4to3d ( &vc[i], &(((point3d*)mesh->vc)[i]) );
    }
  }
}

BSMesh * bsm_NewMesh(void) {
  BSMesh *mesh = NULL;
  PKV_MALLOC(mesh, sizeof(BSMesh));
  if (mesh != NULL) {
     mesh->nv = 0;
     mesh->nhe = 0;
     mesh->nfac = 0;
     mesh->mv = NULL;
     mesh->vc = NULL;
     mesh->mvhei = NULL;
     mesh->mhe = NULL;
     mesh->mfac = NULL;
     mesh->mfhei = NULL;
  }
  return mesh;
}

void bsm_ClearMesh(BSMesh *mesh) {
  if (NULL != mesh) {
     PKV_FREE(mesh->mv);
     PKV_FREE(mesh->vc);
     PKV_FREE(mesh->mvhei);
     PKV_FREE(mesh->mhe);
     PKV_FREE(mesh->mfac);
     PKV_FREE(mesh->mfhei);
     mesh->nv = 0;
     mesh->nhe = 0;
     mesh->nfac = 0;  
  }
}

boolean bsm_ReadMesh(BSMesh *mesh, const char *filename) {
  bsf_UserReaders readers;
  boolean result = false;
  readers.userData = NULL;
  bsf_ClearReaders(&readers);
  readers.BSMeshReader = StoreTheMesh;
  readers.userData = mesh;
  result = bsf_ReadBSFiled(filename, &readers);
  return result;
}

boolean bsm_ReadMeshFromStream(BSMesh *mesh, FILE *stream) {
  return true;
}

boolean bsm_IsMeshEmpty(const BSMesh *mesh) {
  return true;
}

void bsm_WriteMeshInfo(const BSMesh *mesh, FILE *stream) {
  assert(NULL != mesh);
  fprintf(stream, "number of vertices: %d\n", mesh->nv);
  fprintf(stream, "number of halfedges: %d\n", mesh->nhe);
  fprintf(stream, "number of facets: %d\n", mesh->nfac);  
}
