
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>    
#include <math.h>      
#include <malloc.h>
#include <string.h>   
#include <unistd.h> 
#include <sys/times.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkvaria.h"   
#include "pkgeom.h"
#include "pknum.h"
#include "camera.h"

#include "light.h"
#include "plane.h"

/* ///////////////////////////////////////////////////////////////////////// */
#define R   5.0
#define H   0.25

static double xc = 0.0, yc = 0.0;

void SetPlaneCentre ( double x, double y )
{
  xc = x;  yc = y;
} /*SetPlaneCentre*/

void DrawThePlaneWF ( void )
{
  double x0, x1, x, y0, y1, y;

  glColor3f ( 1.0, 0.0, 0.0 );
  glBegin ( GL_LINES );
    x0 = H * ceil  ( (xc - R)/H );
    x1 = H * floor ( (xc + R)/H );
    for ( x = x0; x <= x1; x += H ) {
      y = sqrt ( R*R - (x-xc)*(x-xc) );
      glVertex2d ( x, yc-y );
      glVertex2d ( x, yc+y );
    }
    y0 = H * ceil  ( (yc - R)/H );
    y1 = H * floor ( (yc + R)/H );
    for ( y = y0; y <= y1; y += H ) {
      x = sqrt ( R*R - (y-yc)*(y-yc) );
      glVertex2d ( xc-x, y );
      glVertex2d ( xc+x, y );
    }
  glEnd ();
#undef LL
} /*DrawThePlaneWF*/

/* ////////////////////////////////////////////////////////////////////////// */
#define texImageWidth  128
#define texImageHeight 128
#define texImageW1 64
#define texImageW2 32
#define texImageW3 16
#define texImageW4 8
#define texImageW5 4
#define texImageW6 2
#define texImageW7 1

static GLubyte faceImage[texImageHeight][texImageWidth][4];
static GLubyte faceImage1[texImageW1][texImageW1][4];
static GLubyte faceImage2[texImageW2][texImageW2][4];
static GLubyte faceImage3[texImageW3][texImageW3][4];
static GLubyte faceImage4[texImageW4][texImageW4][4];
static GLubyte faceImage5[texImageW5][texImageW5][4];
static GLubyte faceImage6[texImageW6][texImageW6][4];
static GLubyte faceImage7[texImageW7][texImageW7][4];
/*
static GLubyte Chessboard[2][2][4] =
  {{{140,140,140,255},{120,120,120,255}},
   {{120,120,120,255},{140,140,140,255}}};
*/
GLuint chesstexname;

void FaceTexture ( double x, double y, GLubyte c[4] )
{
  double xx, yy, d;
  GLubyte cv;

  if ( x >= 0.5 )
    xx = x-0.5;
  else
    xx = x;
  if ( y >= 0.5 )
    yy = y-0.5;
  else
    yy = y;

  xx -= 0.25;

  if ( yy >= 0.355 && yy <= 0.38 && fabs(xx) <= 0.15 )
    cv = 150;
  else {
    xx = fabs(xx)-0.09;
    yy -= 0.18;
    d = xx*xx+yy*yy;
    if ( d <= 0.004 )
      cv = 150;
    else
      cv = 255;
  }

  if ( x >= 0.5 )
    cv = 405-cv;
  if ( y >= 0.5 )
    cv = 405-cv;

  c[0] = c[1] = c[2] = cv;
  c[3] = 255;
} /*FaceTexture*/

void makeTexImages ( void )
{
  int i, j, k;

  for ( j = 0; j < texImageHeight; j++ )
    for ( i = 0; i < texImageWidth; i++ )
      FaceTexture ( ((double)i)/texImageWidth, 1.0-((double)j)/texImageHeight,
                    faceImage[j][i] );

                 /* generate textures for MIP mapping */
  for ( j = 0; j < texImageW1; j++ )
    for ( i = 0; i < texImageW1; i++ )
      for ( k = 0; k < 4; k++ )
        faceImage1[i][j][k] = ( (int)faceImage[i+i][j+j][k] +
                                (int)faceImage[i+i+1][j+j][k] +
                                (int)faceImage[i+i][j+j+1][k] +
                                (int)faceImage[i+i+1][j+j+1][k] ) / 4;
  for ( j = 0; j < texImageW2; j++ )
    for ( i = 0; i < texImageW2; i++ )
      for ( k = 0; k < 4; k++ )
        faceImage2[i][j][k] = ( (int)faceImage1[i+i][j+j][k] +
                                (int)faceImage1[i+i+1][j+j][k] +
                                (int)faceImage1[i+i][j+j+1][k] +
                                (int)faceImage1[i+i+1][j+j+1][k] ) / 4;
  for ( j = 0; j < texImageW3; j++ )
    for ( i = 0; i < texImageW3; i++ )
      for ( k = 0; k < 4; k++ )
        faceImage3[i][j][k] = ( (int)faceImage2[i+i][j+j][k] +
                                (int)faceImage2[i+i+1][j+j][k] +
                                (int)faceImage2[i+i][j+j+1][k] +
                                (int)faceImage2[i+i+1][j+j+1][k] ) / 4;
  for ( j = 0; j < texImageW4; j++ )
    for ( i = 0; i < texImageW4; i++ )
      for ( k = 0; k < 4; k++ )
        faceImage4[i][j][k] = ( (int)faceImage3[i+i][j+j][k] +
                                (int)faceImage3[i+i+1][j+j][k] +
                                (int)faceImage3[i+i][j+j+1][k] +
                                (int)faceImage3[i+i+1][j+j+1][k] ) / 4;
} /*makeTexImages*/

void InitPlaneTexture ( void )
{
  makeTexImages ();
  glPixelStorei ( GL_UNPACK_ALIGNMENT, 1 );

  glEnable(GL_TEXTURE_2D);
  glGenTextures ( 1, &chesstexname );
  glBindTexture ( GL_TEXTURE_2D, chesstexname );
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
  glTexImage2D ( GL_TEXTURE_2D, 0, GL_RGBA, texImageWidth, texImageHeight,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage );
  glTexImage2D ( GL_TEXTURE_2D, 1, GL_RGBA, texImageW1, texImageW1,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage1 );
  glTexImage2D ( GL_TEXTURE_2D, 2, GL_RGBA, texImageW2, texImageW2,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage2 );
  glTexImage2D ( GL_TEXTURE_2D, 3, GL_RGBA, texImageW3, texImageW3,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage3 );
  glTexImage2D ( GL_TEXTURE_2D, 4, GL_RGBA, texImageW4, texImageW4,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage4 );
  glTexImage2D ( GL_TEXTURE_2D, 5, GL_RGBA, texImageW5, texImageW5,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage5 );
  glTexImage2D ( GL_TEXTURE_2D, 6, GL_RGBA, texImageW6, texImageW6,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage6 );
  glTexImage2D ( GL_TEXTURE_2D, 7, GL_RGBA, texImageW7, texImageW7,
                 0, GL_RGBA, GL_UNSIGNED_BYTE, faceImage7 );
  glBindTexture ( GL_TEXTURE_2D, 0 );
} /*InitPlaneTexture*/

void BindPlaneTexture ( void )
{
  double tex_cs[4] = {1.0,0.0,0.0,0.0};
  double tex_ct[4] = {0.0,1.0,0.0,0.0};

  glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 100.0 );
  glEnable ( GL_TEXTURE_2D );
  glTexEnvi ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
  glBindTexture ( GL_TEXTURE_2D, chesstexname );
  glTexGeni ( GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
  glTexGendv ( GL_S, GL_OBJECT_PLANE, tex_cs );
  glTexGeni ( GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
  glTexGendv ( GL_T, GL_OBJECT_PLANE, tex_ct );
  glEnable ( GL_TEXTURE_GEN_S );
  glEnable ( GL_TEXTURE_GEN_T );
  glMatrixMode ( GL_TEXTURE );
  glLoadIdentity ();
  glMatrixMode ( GL_MODELVIEW );
} /*BindPlaneTexture*/

void DrawThePlaneSolid ( void )
{
#define NLINES 100
  int     i;
  double  phi;
  GLfloat colour[3] = { 0.5, 0.5, 0.5 };
  GLfloat shadow[4] = { 0.0, 0.0, 0.0, 1.0 };

  glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 10.0 );  
  glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, shadow ); 
  glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, shadow );
  glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, colour );
  glEnable ( GL_COLOR_MATERIAL );
  glColor3fv ( colour );
  glNormal3d ( 0.0, 0.0, 1.0 );
  glBegin ( GL_POLYGON );
    for ( i = 0; i < NLINES; i++ ) {
      glNormal3f ( 0.0, 0.0, 1.0 );
      phi = ((double)i/(double)NLINES)*2.0*PI;
      glVertex3d ( xc+R*cos(phi), yc+R*sin(phi), -0.0001 );
    }
  glEnd ();
  glDisable ( GL_COLOR_MATERIAL );
} /*DrawThePlaneSolid*/

