
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>   
#include <malloc.h>
#include <string.h>

#include <GL/gl.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "pknum.h"  
#include "multibs.h"

#include "sphericon.h"

/* ///////////////////////////////////////////////////////////////////////// */
/* sphericon surface model, made of rational Bezier patches */
int      nhcones = 0;
halfcone hc[2*MAXH*MAXCONES+4];

/* ///////////////////////////////////////////////////////////////////////// */
static void EnterBezierPatch ( trans3d *tr, point4d *cp, point4d *qcp )
{
  int i;

  for ( i = 0; i < 6; i++ )
    Trans3Point4d ( tr, &cp[i], &qcp[i] );
} /*EnterBezierPatch*/

static boolean GenHalfSphericon ( trans3d *tr, int n, int h, boolean dual )
{
#define SQRTH (0.5*SQRT2)
  int      i, n2, ea, eb;
  point4d  cp[6];
  point2d  poly[MAXH*MAXCONES+1];
  point2d  pp;
  vector2d pv;
  double   d, rho, beta, sa, ca, a;
  trans2d  tt;
  boolean  convex;

  convex = h >= 0;
  h = abs ( h );
  rho = cos ( PI/(double)n );
  n2 = n/2;
  if ( !(n & 0x01) && dual ) {
    d = PI/(double)n;
    n2 --;
  }
  else
    d = 0.0;

        /* generate the polygon - section of the sphericon */
  if ( h == 1 ) {
    for ( i = 0; i <= n; i++ )
      SetPoint2d ( &poly[i], cos(2.0*PI*(double)i/(double)n+d),
                             sin(2.0*PI*(double)i/(double)n+d) );
  }
  else {
    if ( n & 0x01 ) {
      if ( convex ) {
        beta = PI/(2.0*(double)n);
        a = 1.0;
      }
      else {
        beta = -PI/(2.0*(double)n);
        a = -1.0-2.0*cos(PI/(double)n);
      }
    }
    else {
      sa = sin ( PI/(double)n );  ca = cos ( PI/(double)n );
      a = (sqrt ( 3.0+2.0*ca*ca ) - 2.0*ca)/3.0;
      if ( convex ) {
        beta = atan2 ( sa, a+2.0*ca );
        a += ca;
      }
      else {
        beta = -atan2 ( sa, a+2.0*ca );
        a = -a-3.0*ca;
      }
    }
    IdentTrans2d ( &tt );
    ShiftTrans2d ( &tt, -a*cos((double)(n+1)*PI/(double)n),
                        -a*sin((double)(n+1)*PI/(double)n) );
    RotTrans2d ( &tt, 2.0*beta/(double)h );
    ShiftTrans2d ( &tt, a*cos((double)(n+1)*PI/(double)n),
                        a*sin((double)(n+1)*PI/(double)n) );
    SetPoint2d ( &poly[0], 1.0, 0.0 );
    for ( i = 1; i < h; i++ )
      TransPoint2d ( &tt, &poly[i-1], &poly[i] );
    IdentTrans2d ( &tt );
    RotTrans2d ( &tt, 2.0*PI/(double)n );
    for ( i = h; i < n*h; i++ )
      TransPoint2d ( &tt, &poly[i-h], &poly[i] );
    if ( dual && !(n & 0x01) ) {
      IdentTrans2d ( &tt );
      RotTrans2d ( &tt, d );
      for ( i = 0; i < n*h; i++ )
        TransPoint2d ( &tt, &poly[i], &poly[i] );
      pkv_RollArray ( poly, sizeof(point2d), n*h, n*h-h/2 );
      if ( !(h & 0x01) )
        dual = false;
    }
    poly[n*h] = poly[0];
    n *= h;
    n2 = n/2;
    if ( !(n & 0x01) && dual )
      n2 --;
  }

  ea = n-1; eb = 0;
  if ( !(n & 0x01) && dual ) {
    SetPoint4d ( &cp[0], poly[0].x, 0.0, 0.0, 1.0 );
    SetPoint4d ( &cp[1], poly[0].x, poly[0].y, 0.0, 1.0 );
    SetPoint4d ( &cp[2], SQRTH*poly[0].x, 0.0, 0.0, SQRTH );
    SetPoint4d ( &cp[3], SQRTH*poly[0].x, SQRTH*poly[0].y, SQRTH*poly[0].y, SQRTH );
    SetPoint4d ( &cp[4], poly[0].x, 0.0, 0.0, 1.0 );
    SetPoint4d ( &cp[5], poly[0].x, 0.0, poly[0].y, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp0 );
    cp[0] = cp[4];
    cp[1] = cp[5];
    cp[3].y = -cp[3].y;
    SetPoint4d ( &cp[5], poly[0].x, -poly[0].y, 0.0, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp1 );
    hc[nhcones].alpha = 0.0;  hc[nhcones].salpha = 0.0;  hc[nhcones].calpha = 1.0;
    hc[nhcones].flat = true;
    hc[nhcones].cyl = false;
    hc[nhcones].ea = hc[nhcones].eb = 0;
    eb ++;
    hc[nhcones].R = hc[nhcones].r = 0.5*poly[0].y;
    nhcones ++;
  }
  for ( i = 0; i < n2; i++ ) {
    MidPoint2d ( &poly[i], &poly[i+1], &pp );
    SubtractPoints2d ( &poly[i+1], &poly[i], &pv );
    SetPoint4d ( &cp[0], poly[i].x, poly[i].y, 0.0, 1.0 );
    SetPoint4d ( &cp[1], poly[i+1].x, poly[i+1].y, 0.0, 1.0 );
    SetPoint4d ( &cp[2], SQRTH*poly[i].x, SQRTH*poly[i].y, SQRTH*poly[i].y, SQRTH );
    SetPoint4d ( &cp[3], SQRTH*poly[i+1].x, SQRTH*poly[i+1].y, SQRTH*poly[i+1].y, SQRTH );
    SetPoint4d ( &cp[4], poly[i].x, 0.0, poly[i].y, 1.0 );
    SetPoint4d ( &cp[5], poly[i+1].x, 0.0, poly[i+1].y, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp0 );
    cp[0] = cp[4];
    cp[1] = cp[5];
    cp[2].y = -cp[2].y;
    cp[3].y = -cp[3].y;
    SetPoint4d ( &cp[4], poly[i].x, -poly[i].y, 0.0, 1.0 );
    SetPoint4d ( &cp[5], poly[i+1].x, -poly[i+1].y, 0.0, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp1 );
    if ( h == 1 ) {
      if ( dual )
        hc[nhcones].alpha = (double)(i+i+2)/(double)n*PI;
      else
        hc[nhcones].alpha = ((double)i+i+1)/(double)n*PI;
    }
    else
      hc[nhcones].alpha = atan2 ( pv.x, pv.y );
    hc[nhcones].salpha = sin ( hc[nhcones].alpha );
    hc[nhcones].calpha = cos ( hc[nhcones].alpha );
    hc[nhcones].flat = false;
    if ( h > 1 ) {
      if ( fabs ( pv.y ) < 1.0e-7 )
        goto cyl;
      else
        hc[nhcones].cyl = false;
    }
    else if ( !(n & 0x01) && i+i+1 == n2 ) {
      if ( ((n & 0x02) && !dual ) ||
           (!(n & 0x02) && dual ) ) {
cyl:
        hc[nhcones].cyl = true;
        hc[nhcones].alpha = 0.5*PI;
        hc[nhcones].salpha = 1.0;
        hc[nhcones].calpha = 0.0;
      }
      else
        hc[nhcones].cyl = false;
    }
    else
      hc[nhcones].cyl = false;
    hc[nhcones].ea = ea --;
    hc[nhcones].eb = eb ++;
    if ( h == 1 )
      hc[nhcones].r = rho*hc[nhcones].salpha;
    else
      hc[nhcones].r = pp.y;
    if ( hc[nhcones].cyl )
      hc[nhcones].R = 0.0;  /* infinity in fact */
    else
      hc[nhcones].R = hc[nhcones].r/hc[nhcones].calpha;
    nhcones ++;
  }
  if ( (n & 0x01) || dual ) {
    SetPoint4d ( &cp[0], poly[n2].x, poly[n2].y, 0.0, 1.0 );
    SetPoint4d ( &cp[1], poly[n2].x, 0.0, 0.0, 1.0 );
    SetPoint4d ( &cp[2], SQRTH*poly[n2].x, SQRTH*poly[n2].y, SQRTH*poly[n2].y, SQRTH );
    SetPoint4d ( &cp[3], SQRTH*poly[n2].x, 0.0, 0.0, SQRTH );
    SetPoint4d ( &cp[4], poly[n2].x, 0.0, poly[n2].y, 1.0 );
    SetPoint4d ( &cp[5], poly[n2].x, 0.0, 0.0, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp0 );
    cp[0] = cp[4];
    cp[1] = cp[5];
    cp[2].y = -cp[2].y;
    SetPoint4d ( &cp[4], poly[i].x, -poly[i].y, 0.0, 1.0 );
    EnterBezierPatch ( tr, cp, hc[nhcones].qcp1 );
    hc[nhcones].alpha = PI;  hc[nhcones].salpha = 0.0;  hc[nhcones].calpha = -1.0;
    hc[nhcones].flat = true;
    hc[nhcones].cyl = false;
    hc[nhcones].ea = hc[nhcones].eb = ea;
    hc[nhcones].R = hc[nhcones].r = 0.5*poly[n2].y;
    nhcones ++;
  }
  return dual;
#undef SQRTH
} /*GenHalfSphericon*/

static void FindNeighbours ( int n, int k, boolean dual, boolean halfdual )
{
  int nb0[4*MAXH*MAXCONES], nb1[4*MAXH*MAXCONES];
  int h, i;

  for ( i = 0; i < nhcones; i++ )
    hc[i].na = hc[i].nb = -1;

  if ( n & 0x01 ) {      /* odd */
    h = (n+1)/2;
    for ( i = h; i < nhcones; i++ ) {
      hc[i].ea = (n+n-1-hc[i].ea-k) % n;
      hc[i].eb = (n+n-1-hc[i].eb-k) % n;
    }
    for ( i = 0; i < h; i++ )
      nb0[i] = nb0[n-1-i] = i+h;
    for ( i = 0; i < n; i++ )
      nb1[(i+n-k) % n] = nb0[i];
    for ( i = 0; i < h; i++ ) {
      hc[i].na = nb1[n-1-i];
      hc[i].nb = nb1[i];
    }
    for ( i = 0; i < h; i++ ) {
      if ( hc[hc[i].na].na == -1 )
        hc[hc[i].na].na = i;
      else
        hc[hc[i].na].nb = i;
      if ( hc[hc[i].nb].na == -1 )
        hc[hc[i].nb].na = i;
      else
        hc[hc[i].nb].nb = i;
    }
    hc[nhcones-1].nb = hc[nhcones-1].na;
  }
  else if ( dual ) {    /* even, dual */
    h = n/2 + 1;
    for ( i = h; i < nhcones; i++ ) {
      hc[i].ea = (n+n-hc[i].ea-k) % n;
      hc[i].eb = (n+n-hc[i].eb-k) % n;
    }
    nb0[0] = h;
    for ( i = 1; i < h; i++ )
      nb0[i] = nb0[n-i] = i + h;
    for ( i = 0; i < n; i++ )
      nb1[(i+n-k) % n] = nb0[i];
    hc[0].na = hc[0].nb = nb1[0];
    for ( i = 1; i < h; i++ ) {
      hc[i].na = nb1[n-i];
      hc[i].nb = nb1[i];
    }
    for ( i = 0; i < h; i++ ) {
      if ( hc[hc[i].na].na == -1 )
        hc[hc[i].na].na = i;
      else
        hc[hc[i].na].nb = i;
      if ( hc[hc[i].nb].na == -1 )
        hc[hc[i].nb].na = i;
      else
        hc[hc[i].nb].nb = i;
    }
    hc[h].nb = hc[h].na;
    hc[nhcones-1].nb = hc[nhcones-1].na;
  }
  else if ( halfdual ) { /* even, halfdual */
    h = n/2;
    for ( i = h+1; i < nhcones; i++ ) {
      hc[i].ea = (n+n-1-hc[i].ea-k) % n;
      hc[i].eb = (n+n-1-hc[i].eb-k) % n;
    }
    for ( i = 0; i < h; i++ )
      nb0[i] = nb0[n-1-i] = i+h+1;
    for ( i = 0; i < n; i++ )
      nb1[(i+n-k) % n] = nb0[i];
    hc[0].na = hc[0].nb = nb1[0];
    for ( i = 1; i < h; i++ ) {
      hc[i].na = nb1[n-i];
      hc[i].nb = nb1[i];
    }
    hc[h].na = hc[h].nb = nb1[h];
    for ( i = 0; i <= h; i++ ) {
      if ( hc[hc[i].na].na == -1 )
        hc[hc[i].na].na = i;
      else
        hc[hc[i].na].nb = i;
      if ( hc[hc[i].nb].na == -1 )
        hc[hc[i].nb].na = i;
      else
        hc[hc[i].nb].nb = i;
    }
  }
  else {                 /* even, nondual */
    h = n/2;
    for ( i = h; i < n; i++ ) {
      hc[i].ea = (n+n-1-hc[i].ea-k) % n;
      hc[i].eb = (n+n-1-hc[i].eb-k) % n;
    }
    for ( i = 0; i < h; i++ )
      nb0[i] = nb0[n-1-i] = i + h;
    for ( i = 0; i < n; i++ )
      nb1[(i+n-k) % n] = nb0[i];
    for ( i = 0; i < h; i++ ) {
      hc[i].na = nb1[n-1-i];
      hc[i].nb = nb1[i];
    }
    for ( i = 0; i < h; i++ ) {
      if ( hc[hc[i].na].na == -1 )
        hc[hc[i].na].na = i;
      else
        hc[hc[i].na].nb = i;
      if ( hc[hc[i].nb].na == -1 )
        hc[hc[i].nb].na = i;
      else
        hc[hc[i].nb].nb = i;
    }
  }
/*
printf ( "\n" );
for ( i = 0; i < nhcones; i++ )
  printf ( "%d: %d (%5.2f,%5.2f),(%5.2f,%5.2f), %d(%5.2f,%5.2f),(%5.2f,%5.2f), %d %d\n",
           i, hc[i].ea, hc[i].qcp0[0].x, hc[i].qcp0[0].y, hc[i].qcp0[1].x, hc[i].qcp0[1].y,
           hc[i].eb, hc[i].qcp1[4].x, hc[i].qcp1[4].y, hc[i].qcp1[5].x, hc[i].qcp1[5].y,
           hc[i].na, hc[i].nb );
*/
} /*FindNeighbours*/

static void ReverseHalfcone ( halfcone *hc )
{
  int i;

  pkv_Exchange ( hc->qcp0, &hc->qcp1[4], 2*sizeof(point4d) );
  pkv_Exchange ( &hc->qcp0[2], &hc->qcp1[2], 2*sizeof(point4d) );
  for ( i = 0; i <= 2; i++ ) {
    pkv_Exchange ( &hc->qcp0[2*i], &hc->qcp0[2*i+1], sizeof(point4d) );
    pkv_Exchange ( &hc->qcp1[2*i], &hc->qcp1[2*i+1], sizeof(point4d) );
  }
  i = hc->ea;  hc->ea = hc->eb;  hc->eb = i;
  i = hc->na;  hc->na = hc->nb;  hc->nb = i;
  hc->calpha = -hc->calpha;
  hc->R = -hc->R;
} /*ReverseHalfcone*/

static void UpdateOrientation ( void )
{
  int i, j, k;

  for ( i = 0; i < nhcones; i++ )
    hc[i].tag = hc[i].flat;
  for ( i = 0; i < nhcones; i++ )
    if ( !hc[i].tag ) {
      hc[i].tag = true;
      k = i;
      j = hc[k].na;
      if ( hc[j].ea != hc[i].ea && hc[j].eb != hc[i].ea )
        ReverseHalfcone ( &hc[i] );
      while ( !hc[j].tag ) {
        hc[j].tag = true;
        if ( hc[j].eb != hc[k].ea )
          ReverseHalfcone ( &hc[j] );
        if ( hc[j].nb != k )
          { hc[j].na = hc[j].nb;  hc[j].nb = k; }
        k = j;
        j = hc[k].na;
      }
      k = i;
      j = hc[k].nb;
      while ( !hc[j].tag ) {
        hc[j].tag = true;
        if ( hc[j].ea != hc[k].eb )
          ReverseHalfcone ( &hc[j] );
        if ( hc[j].na != k )
          { hc[j].nb = hc[j].na;  hc[j].na = k; }
        k = j;
        j = hc[k].nb;
      }
    }
  for ( i = 0; i < nhcones/2; i++ )
    if ( !hc[i].flat ) {
      k = hc[i].nb;
      if ( hc[k].na != i )
        { j = hc[k].na;  hc[k].na = hc[k].nb;  hc[k].nb = j; }
    }
/*
for ( i = 0; i < nhcones; i++ ) {
  printf ( "%d: %d (%5.2f,%5.2f)(%5.2f,%5.2f), %d (%5.2f,%5.2f)(%5.2f,%5.2f), %d, %d\n", i,
           hc[i].ea, hc[i].qcp0[0].x, hc[i].qcp0[0].y, hc[i].qcp0[1].x, hc[i].qcp0[1].y,
           hc[i].eb, hc[i].qcp1[5].x, hc[i].qcp1[5].y, hc[i].qcp1[4].x, hc[i].qcp1[4].y,
           hc[i].na, hc[i].nb );
}
printf ( "\n" );
*/
} /*UpdateOrientation*/

int GenSphericon ( int n, int k, int h, boolean dual, boolean halfdual )
{
  trans3d tr;
  int     i, r;
  boolean hd;

  nhcones = 0;
  IdentTrans3d ( &tr );
  hd = GenHalfSphericon ( &tr, n, h, dual || halfdual );
  RotXTrans3d ( &tr, PI );
  if ( halfdual )
    RotZTrans3d ( &tr, 2.0*PI*((double)k+0.5)/(double)n );
  else
    RotZTrans3d ( &tr, 2.0*PI*(double)k/(double)n );
  dual = GenHalfSphericon ( &tr, n, h, dual );
/*printf ( "dual = %d, halfdual = %d, hd = %d\n", dual, halfdual, hd );*/
  h = abs ( h );
  r = h*k;
  if ( halfdual && h > 1 ) {
    if ( hd )
      r += (h-1)/2;
    else
      r += h/2;
  }
/*printf ( "r = %d\n", r );*/
  FindNeighbours ( n*h, r, dual, hd );
  UpdateOrientation ();
  for ( i = 0; i < nhcones; i++ ) {
    hc[i].phi = 0.5*PI*hc[i].calpha;
    hc[i].sphi = sin ( hc[i].sphi );
    hc[i].cphi = cos ( hc[i].cphi );
    hc[i].lgt = 0.5*PI*hc[i].r;
  }
  return nhcones;
} /*GenSphericon*/

