
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>   
#include <malloc.h>
#include <string.h>

#include <GL/gl.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "pknum.h"  
#include "multibs.h"

#include "sphericon.h"

/* transformation of the sphericon when it is rolled */
trans3d sphtrans1, sphtrans2, sphtrans3;
int     rollhc = 0;      /* facet number */
double  roll_s = 0.0;    /* length */
double  rollphi = 0.0;   /* sphericon phase */
boolean rolldir = true;

point3d hcframe[4];

point3d plframe[3] = {{0.0,0.0,0.0},{1.0,0.0,0.0},{0.0,0.0,1.0}};
point3d spframe[3] = {{1.0,0.0,0.0},{0.0,1.0,0.0},{0.0,0.0,0.0}};

/* ///////////////////////////////////////////////////////////////////////// */
void IdentSphTrans ( void )
{
  vector3d v;

  IdentTrans3d ( &sphtrans1 );
  IdentTrans3d ( &sphtrans2 );
  SetVector3d ( &v, 0.0, 0.0, 1.0 );
  MirrorTrans3d ( &sphtrans2, &v );
  sphtrans3 = sphtrans2;
} /*IdentSphTrans*/

static boolean RollStoPhiT ( double s, double *t )
{
  double phi, sphi, cphi, a, b, c, x1, x2;

  phi = 0.5*PI*s/hc[rollhc].lgt;
  sphi = sin ( fabs(phi) );  cphi = cos ( phi );
  if ( sphi > 0.5*SQRT2 ) {
    a = SQRT2-1.0+(2.0-SQRT2)*sphi;
    b = (SQRT2-2.0)*sphi-SQRT2;
    c = sphi;
  }
  else {
    a = SQRT2-1.0+(2.0-SQRT2)*cphi;
    b = (SQRT2-2.0)*cphi+2.0-SQRT2;
    c = cphi-1.0;
  }
  if ( pkn_SolveSqEqd ( b/(a+a), c/a, &x1, &x2 ) ) {
    if ( x1 < 0.0 || x1 > 1.0 )
      x1 = x2;
    if ( x1 < 0.0 || x1 > 1.0 ) {
printf ( "a" );
      *t = 0.0;
      return false;
    }
    *t = phi >= 0.0 ? x1 : -x1;
    return true;
  }
  else {
printf ( "b" );
    *t = 0.0;
    return false;
  }
} /*RollStoPhiT*/

static void SetSphTrans ( double s, double t )
{
  vector3d v1, v2;
  double   ang, sa, ca;

/*
printf ( "rollhc = %d, r = %6.3f, s = %6.3f, t = %6.3f\n",
          rollhc, hc[rollhc].r, roll_s, t );
*/

  if ( hc[rollhc].cyl ) {
    InterPoint3d ( &hcframe[0], &hcframe[2], s/hc[rollhc].lgt, &plframe[0] );
    AddVector3d ( &plframe[0], &hcframe[1], &plframe[1] );
  }
  else {
    SubtractPoints3d ( &hcframe[0], &hcframe[1], &v1 );
    ang = hc[rollhc].phi*s/hc[rollhc].lgt;
    sa = sin ( ang );  ca = cos ( ang );
    SetVector3d ( &v2, ca*v1.x+sa*v1.y, -sa*v1.x+ca*v1.y, 0.0 );
    AddVector3d ( &hcframe[1], &v2, &plframe[0] );
    if ( hc[rollhc].R > 0.0 )
      plframe[1] = hcframe[1];
    else
      InterPoint3d ( &hcframe[1], &plframe[0], 2.0, &plframe[1] );
  }
  SetPoint3d ( &plframe[2], plframe[0].x, plframe[0].y, 1.0 );

  if ( t >= 0.0 ) {
    mbs_BCHornerP3Rd ( 2, 1, hc[rollhc].qcp1, t, 0.5, &spframe[0] );
    mbs_BCHornerP3Rd ( 2, 1, hc[rollhc].qcp1, t, 0.0, &spframe[1] );
  }
  else {
    mbs_BCHornerP3Rd ( 2, 1, hc[rollhc].qcp0, 1.0+t, 0.5, &spframe[0] );
    mbs_BCHornerP3Rd ( 2, 1, hc[rollhc].qcp0, 1.0+t, 0.0, &spframe[1] );
  }
  SetPoint3d ( &spframe[2], 0.0, 0.0, 0.0 );
  SetRotationToPos3d ( &spframe[0], &spframe[1], &spframe[2],
                       &plframe[0], &plframe[1], &plframe[2], &sphtrans1 );
  sphtrans2 = sphtrans1;
  SetVector3d ( &v1, 0.0, 0.0, 1.0 );
  MirrorTrans3d ( &sphtrans2, &v1 );
  CompITrans3d ( &sphtrans3, &sphtrans2, &sphtrans1 );
} /*SetSphTrans*/

void InitSphTrans ( int fno )
{
  rolldir = true;
  rollhc = fno % nhcones;
  if ( hc[rollhc].flat )
    rollhc = hc[rollhc].nb;
  rollphi = roll_s = 0.0;
  SetPoint3d ( &hcframe[0], 0.0, 0.0, 0.0 );
  if ( hc[rollhc].cyl ) {
    SetPoint3d ( &hcframe[1], 1.0, 0.0, 0.0 );
    SetPoint3d ( &hcframe[2], 0.0, hc[rollhc].lgt, 0.0 );
    SetPoint3d ( &hcframe[3], 0.0, -hc[rollhc].lgt, 0.0 );
  }
  else {
    SetPoint3d ( &hcframe[1], hc[rollhc].R, 0.0, 0.0 );
    SetPoint3d ( &hcframe[2], hc[rollhc].R*(1.0-cos(hc[rollhc].phi)),
                 hc[rollhc].R*sin(hc[rollhc].phi), 0.0 );
    SetPoint3d ( &hcframe[3], hcframe[2].x, -hcframe[2].y, 0.0 );
  }
  SetSphTrans ( 0.0, 0.0 );
} /*InitSphTrans*/

static void FindNextHCFrame ( int next, boolean special )
{
  vector3d v1, v2, v3;
  point3d  nexthcf[4];
  double   sp, cp;

  if ( hc[rollhc].na == hc[rollhc].nb ) {
    if ( (hc[rollhc].ea == hc[next].eb) ^ special )
      goto b;
    else
      goto a;
  }
  if ( next == hc[rollhc].na ) {
a:
    nexthcf[2] = hcframe[3];
    if ( hc[rollhc].cyl )
      v1 = hcframe[1];
    else if ( hc[rollhc].R > 0 )
      SubtractPoints3d ( &hcframe[1], &hcframe[3], &v1 );
    else
      SubtractPoints3d ( &hcframe[3], &hcframe[1], &v1 );
    NormalizeVector3d ( &v1 );
    if ( hc[next].cyl ) {
      nexthcf[1] = v1;
      SetVector3d ( &v2, v1.y, -v1.x, 0.0 );
      AddVector3Md ( &nexthcf[2], &v2, hc[next].lgt, &nexthcf[0] );
      AddVector3Md ( &nexthcf[0], &v2, hc[next].lgt, &nexthcf[3] );
    }
    else {
      AddVector3Md ( &hcframe[3], &v1, hc[next].R, &nexthcf[1] );
      sp = sin ( hc[next].phi );  cp = cos ( hc[next].phi );
      SetVector3d ( &v2, cp*v1.x-sp*v1.y, sp*v1.x+cp*v1.y, 0.0 );
      AddVector3Md ( &nexthcf[1], &v2, -hc[next].R, &nexthcf[0] );
      SetVector3d ( &v3, cp*v2.x-sp*v2.y, sp*v2.x+cp*v2.y, 0.0 );
      AddVector3Md ( &nexthcf[1], &v3, -hc[next].R, &nexthcf[3] );
    }
  }
  else {
b:
    nexthcf[3] = hcframe[2];
    if ( hc[rollhc].cyl )
      v1 = hcframe[1];
    else if ( hc[rollhc].R > 0 )
      SubtractPoints3d ( &hcframe[1], &hcframe[2], &v1 );
    else
      SubtractPoints3d ( &hcframe[2], &hcframe[1], &v1 );
    NormalizeVector3d ( &v1 );
    if ( hc[next].cyl ) {
      nexthcf[1] = v1;
      SetVector3d ( &v2, -v1.y, v1.x, 0.0 );
      AddVector3Md ( &nexthcf[3], &v2, hc[next].lgt, &nexthcf[0] );
      AddVector3Md ( &nexthcf[0], &v2, hc[next].lgt, &nexthcf[2] );
    }
    else {
      AddVector3Md ( &hcframe[2], &v1, hc[next].R, &nexthcf[1] );
      sp = sin ( hc[next].phi );  cp = cos ( hc[next].phi );
      SetVector3d ( &v2, cp*v1.x+sp*v1.y, -sp*v1.x+cp*v1.y, 0.0 );
      AddVector3Md ( &nexthcf[1], &v2, -hc[next].R, &nexthcf[0] );
      SetVector3d ( &v3, cp*v2.x+sp*v2.y, -sp*v2.x+cp*v2.y, 0.0 );
      AddVector3Md ( &nexthcf[1], &v3, -hc[next].R, &nexthcf[2] );
    }
  }
  memcpy ( hcframe, nexthcf, 4*sizeof(point3d) );
  rollhc = next;
} /*FindNextHCFrame*/

void IncreaseSphTrans ( double delta )
{
  int    next;
  double t;

  if ( rolldir )
    roll_s += delta;
  else
    roll_s -= delta;

  if ( roll_s > hc[rollhc].lgt ) {
    do {
      next = hc[rollhc].nb;
      if ( hc[next].flat ) {
        roll_s = hc[rollhc].lgt;
        rolldir = !rolldir;
      }
      else {
        roll_s -= hc[rollhc].lgt;
        FindNextHCFrame ( next, false );
        roll_s -= hc[rollhc].lgt;
      }
    } while ( roll_s > hc[rollhc].lgt );
  }
  else if ( roll_s < -hc[rollhc].lgt ) {
    do {
      next = hc[rollhc].na;
      if ( hc[next].flat ) {
        roll_s = -hc[rollhc].lgt;
        rolldir = !rolldir;
      }
      else {
        roll_s += hc[rollhc].lgt;
        FindNextHCFrame ( next, true );
        roll_s += hc[rollhc].lgt;
      }
    } while ( roll_s < -hc[rollhc].lgt );
  }
  if ( RollStoPhiT ( roll_s, &t ) )
    SetSphTrans ( roll_s, t );
  else
    exit ( 1 );
} /*IncreaseSphTrans*/

void GetSphCentre ( point3d *spc )
{
  SetPoint3d ( spc, sphtrans1.U0.a14, sphtrans1.U0.a24, sphtrans1.U0.a34 );
} /*GetSphCentre*/


