
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdlib.h>   
#include <stdio.h>
#include <math.h>  
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>  
#include <GL/glu.h> 
#include <GL/glx.h>

#include "pkvaria.h" 
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"
#include "multibs.h"
#include "bsmesh.h"
#include "g2blendingd.h"
#include "egholed.h"
#include "mengerc.h"
#include "bsfile.h"
#include "pkrender.h"
#include "xgedit.h"
#include "xgledit.h"

#include "widgets.h"
#include "editor.h"  
#include "editor_bezc.h"
#include "editor_bsc.h"
#include "editor_bezp.h"
#include "editor_bsp.h"
#include "editor_bsm.h"
#include "editor_bsh.h"
#include "pozwalaj.h"


void InitPopup01 ( void )
{
  xge_widget *w;

  w = xge_NewSwitch ( win0, NULL, swP01ACTIVATING, 78, 16, 20+84, 40+30,
                      txtActivating, &activate_readin_objects );
  xge_InitOpenFileMenu ( 0, POPUP01, 400, 232, 20, 40,
                         &filemenu, w, fullpath, shortpath,
                         prevpath, file_filter, filename );
} /*InitPopup01*/

void Popup01CameraReader ( void *usrdata, int ident, CameraRecd *camera )
{
  if ( !camera->parallel ) {  /* now only the perspective camera */
    g00win3D.CPos[4] = g00win3D.CPos[3];  /* save the last projection */
    g00win3D.CPos[3].position = camera->position;
    g00win3D.CPos[3].psi      = camera->psi;
    g00win3D.CPos[3].theta    = camera->theta;
    g00win3D.CPos[3].phi      = camera->phi;
    g00win3D.CPos[3].zmin     = camera->zmin;
    g00win3D.CPos[3].zmax     = camera->zmax;
    g00win3D.CPos[3].vd.persp.f    = camera->vd.persp.f;
    g00win3D.CPos[3].vd.persp.xi0  = camera->vd.persp.xi0;
    g00win3D.CPos[3].vd.persp.eta0 = camera->vd.persp.eta0;
    if ( !CameraSetMappingd ( &g00win3D.CPos[3] ) )
      g00win3D.CPos[3] = g00win3D.CPos[4];  /* restore if wrong data */
  }
} /*Popup01CameraReader*/

void Popup01OpenFile ( void )
{
  geom_object *save_current;

  save_current = current_go;
  if ( !GeomObjectReadFile ( filename, Popup01CameraReader ) )
    DisplayErrorMessage1 ( ErrorMsgCannotOpen, 0 );
  if ( save_current )
    current_go = save_current;
  else
    current_go = first_go;
  SetupObjectSpecificMenus ( current_go );
  if ( xge_IsPopupOn ( popup12 ) )
    UpdateObjectNameList ();
  if ( activate_readin_objects )
    rendered_picture = false;
  xge_RedrawAll ();
} /*Popup01OpenFile*/

int Popup01CallBack ( xge_widget *er, int msg, int key, short x, short y )
{
  switch ( msg ) {
case xgemsg_FILEMENU_OPENFILE:
    switch ( er->id ) {
  case POPUP01:
        xge_RemovePopup ( true );
        Popup01OpenFile ();
        return 1;
  default:
        return false;
    }

case xgemsg_SWITCH_COMMAND:
    switch ( er->id ) {
  case swP01ACTIVATING:
      return 1;
  default:
      return 0;
    }

default:
    return 0;
  }
} /*Popup01CallBack*/

