
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdlib.h>   
#include <stdio.h>
#include <math.h>  
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>  
#include <GL/glu.h> 
#include <GL/glx.h>

#include "pkvaria.h" 
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"
#include "multibs.h"
#include "bsmesh.h"
#include "g2blendingd.h"
#include "egholed.h"
#include "mengerc.h"
#include "bsfile.h"
#include "pkrender.h"
#include "xgedit.h"
#include "xgledit.h"

#include "widgets.h"
#include "editor.h"  
#include "editor_bezc.h"
#include "editor_bsc.h"
#include "editor_bezp.h"
#include "editor_bsp.h"
#include "editor_bsm.h"
#include "editor_bsh.h"
#include "pozwalaj.h"


void InitPopup02 ( void )
{
  xge_widget *w;

  w = xge_NewTextWidget ( win0, NULL, txtP02OBJECTSTOSAVE,
                          96, 16, 20+10, 40+200, txtObjectsToSave );
  w = xge_NewSwitch ( win0, w, swP02ALL, 60, 16, 20+110, 40+200,
                      txtAll, &sw_save_all );
  w = xge_NewSwitch ( win0, w, swP02ACTIVE, 60, 16, 20+190, 40+200,
                      txtActive, &sw_save_active );
  w = xge_NewSwitch ( win0, w, swP02CURRENT, 60, 16, 20+270, 40+200,
                      txtCurrent, &sw_save_current );
  w = xge_NewSwitch ( win0, w, swP02CAMERA, 60, 16, 20+110, 40+220,
                      txtCamera, &sw_save_camera );
  w = xge_NewSwitch ( win0, w, swP02APPEND, 60, 16, 20+10, 40+282-30,
                      txtAppend, &sw_save_append );
  xge_InitSaveFileMenu ( 0, POPUP02, 400, 282, 20, 40,
                         &filemenu, w, fullpath, shortpath,
                         prevpath, file_filter, filename );
} /*InitPopup02*/

boolean WriteOtherData ( bsf_Writer *writer, void *usrdata )
{
  if ( sw_save_camera )
    return bsf_WriteCamera ( writer, &g00win3D.CPos[3], -1 );
  else
    return true;
} /*WriteOtherData*/

void Popup02SaveFile ( void )
{
        /* this procedure may be called also from a button */
        /* in popup00 and popup03 */
  char opt;

  if ( sw_save_active )       opt = GO_WRITE_ACTIVE;
  else if ( sw_save_current ) opt = GO_WRITE_CURRENT;
  else                        opt = GO_WRITE_ALL;
  if ( !GeomObjectWriteFile ( filename, opt, WriteOtherData,
                              NULL, sw_save_append ) )
    DisplayErrorMessage1 ( ErrorMsgCannotSave, 0 );
} /*Popup02SaveFile*/

int Popup02CallBack ( xge_widget *er, int msg, int key, short x, short y )
{
  switch ( msg ) {
case xgemsg_FILEMENU_SAVEFILE:
    switch ( er->id ) {
  case POPUP02:
      if ( xge_FilenameCorrect ( filename ) ) {
        xge_RemovePopup ( true );
        Popup02SaveFile ();
      }
      else
        DisplayErrorMessage1 ( ErrorMsgIncorrectFilename, -1 );
      return 1;
  default:
      return 0;
    }

case xgemsg_SWITCH_COMMAND:
    switch ( er->id ) {
  case swP02ALL:
      if ( sw_save_all )
        sw_save_active = sw_save_current = false;
      else
        { sw_save_active = true;  sw_save_current = false; }
      goto redraw_popup;
  case swP02ACTIVE:
      if ( sw_save_active )
        sw_save_current = sw_save_all = false;
      else
        { sw_save_current = true;  sw_save_all = false; }
      goto redraw_popup;
  case swP02CURRENT:
      if ( sw_save_current )
        sw_save_all = sw_save_active = false;
      else
        { sw_save_all = true;  sw_save_active = false; }
      goto redraw_popup;
redraw_popup:
      xge_SetClipping ( filemenu.er2 );
      filemenu.er2->redraw ( filemenu.er2, true );
      return 1;
  case swP02CAMERA:
      return 1;
  case swP02APPEND:
      return 1;
  default:
      return 0;
    }

default:
    return 0;
  }
} /*Popup02CallBack*/

