
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkgeom.h"
#include "xgedit.h"  
#include "glopt.h"
#include "xgledit.h"

#include "gl2Dwin.h"  

#ifdef XGLE_OPENGL_NEW

#include "gl2D3Dwin.glsl.c"

/* ////////////////////////////////////////////////////////////////////////// */
my_gl_programs prog;

void CreateGLPrograms ( void )
{
        /* Compile the shaders and link the programs */
  prog.sh0[0] = xgle_CompileShaderStrings ( GL_VERTEX_SHADER, 3, _shader_v0 );
  prog.sh0[1] = xgle_CompileShaderStrings ( GL_GEOMETRY_SHADER, 4, _shader_g0 );
  prog.sh0[2] = xgle_CompileShaderStrings ( GL_FRAGMENT_SHADER, 2, _shader_f );
  prog.p0 = xgle_LinkProgram ( 3, prog.sh0 );
  prog.sh1[0] = xgle_CompileShaderStrings ( GL_VERTEX_SHADER, 4, _shader_v1 );
  prog.sh1[1] = prog.sh0[2];
  prog.p1 = xgle_LinkProgram ( 2, prog.sh1 );
        /* get the uniform variables to describe the transformations */
          /* model transfrmation */
  prog.trbi[0] = glGetUniformBlockIndex ( prog.p0, UniformTBNames[0] );
  prog.trbi[1] = glGetUniformBlockIndex ( prog.p1, UniformTBNames[0] );
  glGetActiveUniformBlockiv ( prog.p0, prog.trbi[0],
                              GL_UNIFORM_BLOCK_DATA_SIZE, &prog.trbsize );
  glGetUniformIndices ( prog.p0, 2, &UniformTBNames[1], prog.trbind );
  glGetActiveUniformsiv ( prog.p0, 2, prog.trbind, GL_UNIFORM_OFFSET, prog.trbofs );
            /* create and attach the buffer */
  glGenBuffers ( 1, &prog.trbbuf );
  glBindBuffer ( GL_UNIFORM_BUFFER, prog.trbbuf );
  glBufferData ( GL_UNIFORM_BUFFER, prog.trbsize, NULL, GL_DYNAMIC_DRAW );
  glUniformBlockBinding ( prog.p0, prog.trbi[0], BINDPOINT_TRANS );
  glUniformBlockBinding ( prog.p1, prog.trbi[1], BINDPOINT_TRANS );
  glBindBufferBase ( GL_UNIFORM_BUFFER, BINDPOINT_TRANS, prog.trbbuf );
          /* view and projection */
  prog.projbi[0] = glGetUniformBlockIndex ( prog.p0, UniformPBNames[0] );
  prog.projbi[1] = glGetUniformBlockIndex ( prog.p1, UniformPBNames[0] );
  glGetActiveUniformBlockiv ( prog.p0, prog.projbi[0],
                              GL_UNIFORM_BLOCK_DATA_SIZE, &prog.projbsize );
  glGetUniformIndices ( prog.p0, 4, &UniformPBNames[1], prog.projbind );
  glGetActiveUniformsiv ( prog.p0, 4, prog.projbind, GL_UNIFORM_OFFSET,
                          prog.projbofs );
            /* create and attach the buffer */
  glGenBuffers ( 1, &prog.projbbuf );
  glBindBuffer ( GL_UNIFORM_BUFFER, prog.projbbuf );
  glBufferData ( GL_UNIFORM_BUFFER, prog.projbsize, NULL, GL_DYNAMIC_DRAW );
  glUniformBlockBinding ( prog.p0, prog.projbi[0], BINDPOINT_PROJ );
  glUniformBlockBinding ( prog.p1, prog.projbi[1], BINDPOINT_PROJ );
  glBindBufferBase ( GL_UNIFORM_BUFFER, BINDPOINT_PROJ, prog.projbbuf );

  xgle_ExitOnGLError ( "Program preparation failed" );
} /*CreateGLPrograms*/

/* ////////////////////////////////////////////////////////////////////////// */
triangle_va tva;

void CreateVertexArray ( int nv )
{
  glGenVertexArrays ( 1, &tva.vao );
  glBindVertexArray ( tva.vao );
  glGenBuffers ( 2, tva.vbo );
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER, nv*sizeof(point2f), NULL, GL_DYNAMIC_DRAW );
  glEnableVertexAttribArray ( VERTEX_POS );
  glVertexAttribPointer ( VERTEX_POS, 2, GL_FLOAT, GL_FALSE,
                          sizeof(point2f), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER, nv, NULL, GL_DYNAMIC_DRAW );
  glEnableVertexAttribArray ( VERTEX_MARKING );
  glVertexAttribIPointer ( VERTEX_MARKING, 1, GL_BYTE, 1, (GLvoid*)0 );
  glBindVertexArray ( 0 );

  xgle_ExitOnGLError ( "Vertex array preparation failed" );
} /*CreateVertexArray*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitGLMachinery ( void )
{
  CreateGLPrograms ();
  CreateVertexArray ( 3 );
} /*InitGLMachinery*/

void DestroyGLMachinery ( void )
{
} /*DestroyGLMachinery*/

/* ////////////////////////////////////////////////////////////////////////// */
void LoadModelMatrix ( GLfloat model[16] )
{
  GLfloat m[16];

  glBindBuffer ( GL_UNIFORM_BUFFER, prog.trbbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.trbofs[0], 16*sizeof(GLfloat), model );
  pkv_TransposeMatrixf ( 4, 4, 4, model, 4, m );
  pkn_GaussInvertMatrixf ( 4, m );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.trbofs[1], 16*sizeof(GLfloat), m );
} /*LoadModelMatrix*/

void LoadProjectionMatrices ( GLfloat view[16], GLfloat projection[16],
                              GLfloat eyepos[4] )
{
  GLfloat m[16];

  glBindBuffer ( GL_UNIFORM_BUFFER, prog.projbbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.projbofs[0], 16*sizeof(GLfloat), view );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.projbofs[1], 16*sizeof(GLfloat),
                    projection );
        /* OpenGL matrices are stored columnwise; hence, the order of the factors */
        /* is reversed, as we have transpositions here */
  pkn_MultMatrixf ( 4, 4, 4, view, 4, 4, projection, 4, m );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.projbofs[2], 16*sizeof(GLfloat), m );
  glBufferSubData ( GL_UNIFORM_BUFFER, prog.projbofs[3], 4*sizeof(GLfloat), eyepos );
} /*LoadProjectionMatrices*/

void LoadVertices ( int nv, point2f *v )
{
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[0] );
  glBufferSubData ( GL_ARRAY_BUFFER, 0, nv*sizeof(point2f), v );
  glBindBuffer ( GL_ARRAY_BUFFER, 0 );
} /*LoadVertices*/

void LoadVertex ( int vi, point2f *v )
{
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[0] );
  glBufferSubData ( GL_ARRAY_BUFFER, vi*sizeof(point2f), sizeof(point2f), v );
  glBindBuffer ( GL_ARRAY_BUFFER, 0 );
} /*LoadVertex*/

void LoadVMarkings ( int nv, byte *mk )
{
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[1] );
  glBufferSubData ( GL_ARRAY_BUFFER, 0, nv, mk );
  glBindBuffer ( GL_ARRAY_BUFFER, 0 );
} /*LoadVMarkings*/

void LoadVMarking ( int vi, byte mk )
{
  glBindBuffer ( GL_ARRAY_BUFFER, tva.vbo[1] );
  glBufferSubData ( GL_ARRAY_BUFFER, vi, 1, &mk );
  glBindBuffer ( GL_ARRAY_BUFFER, 0 );
} /*LoadVMarking*/

#endif

