
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkgeom.h"
#include "xgedit.h"

#include "widgets.h"

xge_fourww       fourww;
xge_string_ed    stred;
xge_int_widget   intwidg0, intwidg1;
quaterniond      q;
trans3d          qtr;
xge_quatrotballd qball;
char             text[21] = "cosik";
float            dialpos1, dialpos2;
int              intvalue0 = 1, intvalue1 = -1;

xge_listbox    filelist, dirlist;
xge_widget     *mainmenu, *fourwwedr, *popupmenu, *colourpopup, *colourwdg;

float sl2f[2] = {0.0,1.0};

xge_colourmenuf cmenu;
float           colour[3] = {1.0,1.0,1.0};

xge_filemenu    fm;
char            fullpath[XGE_MAX_PATH_LENGTH+1],
                shortpath[XGE_SHORT_PATH_LENGTH+1],
                prevpath[XGE_MAX_PATH_LENGTH+1],
                filter[XGE_MAX_FILENAME_LENGTH+1] = "*.c",
                filename[XGE_MAX_FILENAME_LENGTH+1];

Cursor cursor[XC_num_glyphs/2];

/* ///////////////////////////////////////////////////////////////////////// */
void RysujOkno ( xge_widget *er, boolean onscreen )
{
  xgeSetForeground ( xgec_Blue6 );
  xgeFillRectangle ( er->w-2, er->h-2, er->x+1, er->y+1 );
  xgeSetForeground ( xgec_CornflowerBlue );
  xgeDrawRectangle ( er->w-1, er->h-1, er->x, er->y );
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*RysujOkno*/

boolean OknoMsg ( xge_widget *er, int msg, int key, short x, short y )
{
/*  printf ( "%d, %d, %d, %d, %d\n", er->id, msg, key, x, y ); */
  switch ( msg ) {
case xgemsg_ENTERING:
    xge_SetCurrentWindowCursor ( xgeCURSOR_ARROW );
    break;
case xgemsg_EXITING:
    xge_SetCurrentWindowCursor ( xgeCURSOR_DEFAULT );
    break;
default:
    break;
  }
  return true;
} /*OknoMsg*/

boolean MyMenuMsg ( xge_widget *er, int msg, int key, short x, short y )
{
  switch ( msg ) {
case xgemsg_ENTERING:
    if ( intvalue1 >= 0 )
      xge_SetCurrentWindowCursor ( cursor[intvalue1] );
    break;
case xgemsg_EXITING:
    if ( intvalue1 >= 0 )
      xge_SetCurrentWindowCursor ( None );
    break;
default:
    break;    
  }
  return xge_MenuMsg ( er, msg, key, x, y );
} /*MyMenuMsg*/

/* ///////////////////////////////////////////////////////////////////////// */
int CallBack ( xge_widget *er, int msg, int key, short x, short y )
{
  if ( er ) {
/*    printf ( "%d, %d, %d, %d, %d\n", er->id, msg, key, x, y ); */
    switch ( msg ) {
case xgemsg_BUTTON_COMMAND:
      switch ( er->id ) {
  case btnQuit:
        xge_done = 1;
        break;

  case btnOpen:
        xge_AddOpenFilePopup ( &fm );
        break;

  case btnSave:
        xge_AddSaveFilePopup ( &fm );
        break;

  case btnColour:
        xge_AddPopup ( cmenu.er, true, true );
        break;

  default:
        break;
      }
      break;

case xgemsg_INT_WIDGET_COMMAND:
      switch ( er->id ) {
  case intw1:
        intvalue0 = key;
        printf ( "%d\n", intvalue0 );
        break;

  case intw2:
        intvalue1 = key;
        if ( key < 0 )
          xge_SetCurrentWindowCursor ( None );
        else
          xge_SetCurrentWindowCursor ( cursor[key] );
        break;

  default:
        break;
      }
      break;

case xgemsg_TEXT_EDIT_VERIFY:
      return 1;

case xgemsg_TEXT_EDIT_ENTER:
      return 1;

case xgemsg_QUATROTBALL_COMMAND:
      return 1;

case xgemsg_FILEMENU_CHDIR:
      printf ("chdir\n" );
      break;

case xgemsg_FILEMENU_OPENFILE:
      printf ( "open\n" );
      xge_RemovePopup ( true );
      break;

case xgemsg_FILEMENU_SAVEFILE:
      printf ( "save\n" );
      xge_RemovePopup ( true );
      break;

case xgemsg_FILEMENU_CANCEL:
      printf ( "cancel\n" );
      break;

default:
      break;
    }
  }
  else {
/*    printf ( "%d, %d, %d, %d\n", msg, key, x, y ); */
    switch ( msg ) {
case xgemsg_KEY:
      switch ( key ) {
  case 'Q': case 'q':
        xge_done = 1;
        break;
  default:
        break;
      }
      break;

case xgemsg_RESIZE:
      fourww.er->msgproc ( fourww.er, xgemsg_RESIZE, 0,
                           (short)(xge_current_width-110), xge_current_height );
      mainmenu->msgproc ( mainmenu, xgemsg_RESIZE, 0,
                          110, xge_current_height );
      xge_Redraw ();
      break;

default:
      break;
    }
  }
  return 0;
} /*CallBack*/

/* ///////////////////////////////////////////////////////////////////////// */
static char b0[]        = "Quit";
static char dialtext1[] = "aside";
static char dialtext2[] = "below";
static char intwtext0[] = "int widget";
static char intwtext1[] = "cursor";
static char b8[]        = "Colour";
static char qbtext[]    = "q. ball";

void PrepareCursors ( void )
{
  int i;

  for ( i = 0; i < XC_num_glyphs; i += 2 )
    cursor[i/2] = XCreateFontCursor ( xgedisplay, i );
} /*PrepareCursors*/

void init_edwin ( void  )
{
  xge_widget *rp;

  pkv_InitScratchMem ( 655360 );
  PrepareCursors ();
        /* the main menu */
  rp = xge_NewButton ( 0, NULL, btnQuit, 60, 18, 0, 0, b0 );
  rp = xge_NewButton ( 0, rp, btnOpen, 60, 18, 0, 20, xgetxt_Open );
  rp = xge_NewButton ( 0, rp, btnSave, 60, 18, 0, 40, xgetxt_Save );
  rp = xge_NewStringEd ( 0, rp, strEdText, 100, 18, 0, 60, 60, text, &stred );
  rp = xge_NewDialf ( 0, rp, dial1, 110, 40, 0, 80, dialtext1, &dialpos1 );
  rp = xge_NewDialf ( 0, rp, dial2, 40, 60, 0, 122, dialtext2, &dialpos2 );
  rp = xge_NewIntWidget ( 0, rp, intw1, 110, 18, 0, 184, 0, 20,
                          &intwidg0, intwtext0, &intvalue0 );
  rp = xge_NewIntWidget ( 0, rp, intw2, 110, 18, 0, 204, -1, (XC_num_glyphs/2)-1,
                          &intwidg1, intwtext1, &intvalue1 );
  rp = xge_NewVSlidebar2f ( 0, rp, slide, 10, 109, 0, 224, sl2f );
  rp = xge_NewButton ( 0, rp, btnColour, 60, 18, 0, 340, b8 );
  rp = xge_NewQuatRotBalld ( 0, rp, qball1, 81, 100, 14, 224,
                             &qball, &q, &qtr, qbtext );
  mainmenu = xge_NewMenu ( 0, NULL, menu1, 110, xge_HEIGHT, 0, 0, rp );
  mainmenu->msgproc = MyMenuMsg;

        /* the four window widget */
  rp = xge_NewWidget ( 0, NULL, 0, xge_WIDTH-110, xge_HEIGHT, 110, 0,
                       NULL, NULL, OknoMsg, RysujOkno );
  rp = xge_NewWidget ( 0, rp, 1, xge_WIDTH-110, xge_HEIGHT, 110, 0,
                       NULL, NULL, OknoMsg, RysujOkno );
  rp = xge_NewWidget ( 0, rp, 2, xge_WIDTH-110, xge_HEIGHT, 110, 0,
                       NULL, NULL, OknoMsg, RysujOkno );
  rp = xge_NewWidget ( 0, rp, 3, xge_WIDTH-110, xge_HEIGHT, 110, 0,
                       NULL, NULL, OknoMsg, RysujOkno );
  fourwwedr = xge_NewFourWW ( 0, mainmenu, 2, xge_WIDTH-110, xge_HEIGHT, 110, 0,
                              rp, &fourww );

        /* the file popup menus */
  xge_InitOpenFileMenu ( 0, 2048, 400, 232, 20, 40,
                         &fm, NULL, fullpath, shortpath,
                         prevpath, filter, filename );
  xge_InitSaveFileMenu ( 0, 2048, 400, 232, 20, 40,
                         &fm, NULL, fullpath, shortpath,
                         prevpath, filter, filename );

        /* the colour popup */
  xge_InitColourMenuf ( 0, 10, 40, 40, &cmenu, colour );

  xge_SetWinEdRect ( fourwwedr );
  xge_Redraw ();
} /*init_edwin*/

void destroy_edwin ( void )
{
  pkv_DestroyScratchMem ();
} /*destroy_edwin*/

int main ( int argc, char *argv[] )
{
  setvbuf ( stdout, NULL, _IONBF, 0 );
  if ( !xge_Init ( argc, argv, CallBack, NULL ) )
    exit ( 1 );
  init_edwin ();
  xge_MessageLoop ();
  destroy_edwin ();  
  xge_Cleanup ();  
  exit ( 0 );
} /*main*/   

