
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadKnotSequenced ( pks_scanner *sc, int maxlastknot,
                                int *lastknot, double *knots, boolean *closed )
{
  int i;

  if ( closed )
    *closed = false;
  if ( sc->nextsymbol == PKS_SYMB_IDENT &&
       sc->id_space == ID_SPACE_BSFILE &&
       sc->nextinteger == BSF_KEY_CLOSED ) {
    if ( closed )
      *closed = true;
    sc->GetNextSymbol ( sc );
  }
  switch ( sc->nextsymbol ) {
case PKS_SYMB_IDENT:
    if ( sc->id_space == ID_SPACE_BSFILE &&
         sc->nextinteger == BSF_KEY_UNIFORM ) {
      sc->GetNextSymbol ( sc );
      if ( sc->nextsymbol != PKS_SYMB_INTEGER || sc->nextinteger < 1 )
        return false;
      if ( sc->nextinteger > maxlastknot ) 
        return false;
      *lastknot = sc->nextinteger;
      for ( i = 0; i <= sc->nextinteger; i++ )
        knots[i] = (double)i;
      sc->GetNextSymbol ( sc );
      return true;
    }
    else
      return false;

case PKS_SYMB_LBRACE:
    sc->GetNextSymbol ( sc );
    i = 0;
    for (;;) {
      if ( !bsf_ReadDoubleNumber ( sc, &knots[i] ) )
        return false;
      i ++;
      if ( sc->nextsymbol == PKS_SYMB_COMMA )
        sc->GetNextSymbol ( sc );
      else
        break;
    }
    if ( sc->nextsymbol != PKS_SYMB_RBRACE )
      return false;
    sc->GetNextSymbol ( sc );
    i--;
    if ( i < 1 )
      return false;
    *lastknot = i;
        /* the knot sequence must be nondecreasing */
    while ( i > 0 )
      if ( knots[i] >= knots[i-1] )
        i--;
      else
        return false;
    return true;

default:
    return false;
  }
} /*bsf_ReadKnotSequenced*/

