
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadBSplineCurve4d ( bsf_Reader *reader,
                                 int maxdeg, int maxlastknot, int maxncpoints,
                                 int *deg, int *lastknot, double *knots,
                                 boolean *closed, point4d *cpoints,
                                 int *spdimen, boolean *rational,
                                 char *name, int *ident )
{
  boolean _name, degree, knotsu, c_points, dimen, _id;
  int     ncpoints, dim, cpdimen;

  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_BSCURVE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

        /* nothing has been read in yet */
  _name = degree = knotsu = c_points = *rational = dimen = _id = false;
  ncpoints = 0;
  if ( name )
    *name = 0;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_DEGREE:
        if ( degree )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadCurveDegree ( &reader->sc, maxdeg, deg ) )
          goto failure;
        degree = true;
        break;

  case BSF_KEY_DIM:
        if ( dimen )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadSpaceDim ( &reader->sc, 4, &dim ) )
          goto failure;
        dimen = true;
        break;

  case BSF_KEY_KNOTS:
        if ( knotsu )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadKnotSequenced ( &reader->sc, maxlastknot,
                                      lastknot, knots, closed ) )
          goto failure;
        knotsu = true;
        break;

  case BSF_KEY_CPOINTS:
        if ( c_points )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        ncpoints = bsf_ReadPointsd ( &reader->sc, 4, maxncpoints,
                                     &cpoints[0].x, &cpdimen );
        c_points = true;
        if ( !ncpoints )
          goto failure;
        break;

  case BSF_KEY_RATIONAL:
        reader->sc.GetNextSymbol ( &reader->sc );
        *rational = true;
        break;

        /* optional attributes */
  case BSF_KEY_CPOINTSMK:
        if ( !_bsf_ReadCPMark ( reader, maxncpoints ) )
          goto failure;
        break;

  case BSF_KEY_COLOR:
  case BSF_KEY_COLOUR:
        if ( !_bsf_ReadColour ( reader ) )
          goto failure;
        break;

  default:
        goto failure;  /* anything else is forbidden */
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( (*rational && dim < cpdimen-1) || (!*rational && dim < cpdimen) )
        goto failure;  
      *spdimen = dim;
      if ( degree && knotsu && c_points && ncpoints == *lastknot-*deg )
        return true;
      else
        goto failure;

default:
      goto failure;  /* anything else is forbidden */
    }
  }

failure:
  return false;
} /*bsf_ReadBSplineCurve4d*/

