
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_WriteBSplinePatchd ( bsf_Writer *writer,
                                 int spdimen, int cpdimen, boolean rational,
                                 int udeg, int lastknotu, const double *knotsu,
                                 int vdeg, int lastknotv, const double *knotsv,
                                 boolean closed_u, boolean closed_v,
                                 int pitch, const double *cpoints,
                                 const char *name, int ident,
                                 bsf_WriteAttr_fptr WriteAttr, void *userData )
{
  int sci;

  sci = writer->current_indentation;
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s", "%B-spline patch" );
  BSFeol
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s {", bsf_keyword[BSF_KEY_BSPATCH] );
  BSFeol
  writer->current_indentation += 2;
  if ( name && *name ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s \"%s\"",
              bsf_keyword[BSF_KEY_NAME], name );
    BSFeol
  }
  bsf_WriteIdent ( writer, ident );
  bsf_WriteSpaceDim ( writer, spdimen, rational );
  bsf_WritePatchDegree ( writer, udeg, vdeg );
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s", bsf_keyword[BSF_KEY_KNOTS_U] );
  bsf_WriteKnotSequenced ( writer, lastknotu, knotsu, closed_u );
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s", bsf_keyword[BSF_KEY_KNOTS_V] );
  bsf_WriteKnotSequenced ( writer, lastknotv, knotsv, closed_v );
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s ", bsf_keyword[BSF_KEY_CPOINTS] );
  writer->current_indentation += 2;
  bsf_WritePointsd ( writer,
                     cpdimen, lastknotu-udeg, lastknotv-vdeg, pitch, cpoints );
  writer->current_indentation -= 2;
  if ( WriteAttr )
    WriteAttr ( writer, userData );
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation = sci;
  return true;
} /*bsf_WriteBSplinePatchd*/

