
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_WriteBSplineHoled ( bsf_Writer *writer,
                                int spdimen, int cpdimen, boolean rational,
                                int hole_k, const double *knots,
                                const point2d *domain_cp, const double *hole_cp,
                                const char *name, int ident,
                                bsf_WriteAttr_fptr WriteAttr, void *userData )
{
  int i;
  int sci;

  sci = writer->current_indentation;
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s", "%Hole in a piecewise bicubic B-spline surface" );
  BSFeol
  BSFwci
  writer->current_length += fprintf ( writer->output,
          "%s {", bsf_keyword[BSF_KEY_BSHOLE] );
  BSFeol
  writer->current_indentation += 2;
  if ( name && *name ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s \"%s\"",
              bsf_keyword[BSF_KEY_NAME], name );
    BSFeol
  }
  bsf_WriteSpaceDim ( writer, spdimen, rational );
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            bsf_keyword[BSF_KEY_SIDES], hole_k );
  BSFeol
  for ( i = 0; i < hole_k; i++ ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s",
              bsf_keyword[BSF_KEY_KNOTS] );
    writer->current_indentation += 2;
    bsf_WriteKnotSequenced ( writer, 10, &knots[11*i], false );
    writer->current_indentation -= 2;
    BSFeol
  }
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %s ",
            bsf_keyword[BSF_KEY_DOMAIN], bsf_keyword[BSF_KEY_CPOINTS] );
  writer->current_indentation += 2;
  bsf_WritePointsd ( writer, 2, 1, 12*hole_k+1, 0, &domain_cp[0].x );
  writer->current_indentation -= 2;
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s ",
            bsf_keyword[BSF_KEY_CPOINTS] );
  BSFeol
  writer->current_indentation += 2;
  bsf_WritePointsd ( writer, cpdimen, 1, 12*hole_k+1, 0, hole_cp );
  writer->current_indentation -= 2;
  if ( WriteAttr )
    WriteAttr ( writer, userData );
  writer->current_indentation = sci;
  BSFwci
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  return true;
} /*bsf_WriteBSplineHoled*/

