
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2013, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_WriteCamera ( bsf_Writer *writer, CameraRecd *camera, int ident )
{
  int sci;

  sci = writer->current_indentation;
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s {",
            bsf_keyword[BSF_KEY_CAMERA] );
  BSFeol
  writer->current_indentation += 2;
  BSFwci
  bsf_WriteIdent ( writer, ident );
  writer->current_length += fprintf ( writer->output, "%s {",
            bsf_keyword[BSF_KEY_FRAME] );
  writer->current_length += fprintf ( writer->output, "%d, %d, %d, %d}",
            camera->width, camera->height, camera->xmin, camera->ymin );
  BSFeol
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s ",
            bsf_keyword[BSF_KEY_POSITION] );
  bsf_WritePointd ( writer, 3, &camera->position.x );
  BSFeol
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s {",
            bsf_keyword[BSF_KEY_EULERANGLES] );
  bsf_WriteDoubleNumber ( writer, camera->psi );
  writer->current_length += fprintf ( writer->output, ", " );
  bsf_WriteDoubleNumber ( writer, camera->theta );
  writer->current_length += fprintf ( writer->output, ", " );
  bsf_WriteDoubleNumber ( writer, camera->phi );
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  if ( camera->zmin > 0.0 && camera->zmax > camera->zmin ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s { ",
              bsf_keyword[BSF_KEY_DEPTH] );
    bsf_WriteDoubleNumber ( writer, camera->zmin );
    writer->current_length += fprintf ( writer->output, ", " );
    bsf_WriteDoubleNumber ( writer, camera->zmax );
    writer->current_length += fprintf ( writer->output, " }" );
    BSFeol
  }
  BSFwci
  if ( camera->parallel ) {  /* parallel projection camera */
    writer->current_length += fprintf ( writer->output, "%s { ",
              bsf_keyword[BSF_KEY_PARALLEL] );
    bsf_WriteDoubleNumber ( writer, camera->vd.para.wdt );
    writer->current_length += fprintf ( writer->output, ", " );
    bsf_WriteDoubleNumber ( writer, camera->vd.para.hgh );
    writer->current_length += fprintf ( writer->output, ", " );
    bsf_WriteDoubleNumber ( writer, camera->vd.para.diag );
    writer->current_length += fprintf ( writer->output, ", %d}",
                                    (int)camera->vd.para.dim_case );
  }
  else {  /* perspective projection camera */
    writer->current_length += fprintf ( writer->output, "%s { ",
              bsf_keyword[BSF_KEY_PERSPECTIVE] );
    bsf_WriteDoubleNumber ( writer, camera->vd.persp.f );
    writer->current_length += fprintf ( writer->output, ", " );
    bsf_WriteDoubleNumber ( writer, camera->vd.persp.xi0 );
    writer->current_length += fprintf ( writer->output, ", " );
    bsf_WriteDoubleNumber ( writer, camera->vd.persp.eta0 );
    writer->current_length += fprintf ( writer->output, "}" );
  }
  BSFeol
  writer->current_indentation = sci;
  BSFwci
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  return true;
} /*bsf_WriteCamera*/

