
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "bsmesh.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

/* The procedure bsf_ReadIdentifiers reads the entire .bs data file, scanning */
/* it to find all identifiers (ident n). It is supposed to be used by         */
/* applications before appending data to the .bs file in order to make all    */
/* identifiers unique through the .bs file. It is legal not to assign         */
/* identifiers to the objects (call the writing procedures with the parameter */
/* ident equal to -1) and ignore these identifiers. The identifiers are       */
/* intended to carry the information about dependencies among the objects.    */

/* the return value is true if the specified file is empty or it has been     */
/* successfully read. For each identifier the application procedure is called */
/* and it should make a list of the identifiers. Then, generating the         */
/* identifiers of objects to append the application ought to avoid the        */
/* numbers from the list.                                                     */
boolean bsf_ReadIdentifiers ( const char *filename, bsf_Reader *reader,
                              boolean (*readident)( bsf_Reader *reader,
                                                    int objtype, int ident ) )
{
  void *sp;
  int  objtype;

  sp = pkv_GetScratchMemTop ();
  if ( !_bsf_OpenInputFile ( filename, reader ) )
    return false;
  objtype = -1;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_BCURVE:
  case BSF_KEY_BPATCH:
  case BSF_KEY_BSCURVE:
  case BSF_KEY_BSPATCH:
  case BSF_KEY_BSMESH:
  case BSF_KEY_BSHOLE:
        objtype = reader->sc.nextsymbol;
        break;
  case BSF_KEY_IDENT:
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
          goto failure;
        readident ( reader, objtype, reader->sc.nextinteger );
        break;
  default:
        break;
      }
      break;
case PKS_SYMB_ERROR:
      goto failure;
case PKS_SYMB_EOF:
      goto finish;
default:
      break;
    }
    reader->sc.GetNextSymbol ( &reader->sc );
  }
finish:
  pks_ShutDownPreprocessor ( &reader->sc );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pks_ShutDownPreprocessor ( &reader->sc );
  pkv_SetScratchMemTop ( sp );
  return false;
} /*bsf_ReadIdentifiers*/

