
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "bsmesh.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

/* dependencies are specified by a keyword followed by a comma-separated   */
/* list of integer object identifiers in braces. The objects with these    */
/* identifiers ought to be present in the file, but the library procedures */
/* do not check it - it is left to be dealt with by the application.       */

boolean bsf_ReadDependencies ( bsf_Reader *reader )
{
  void *sp;
  int  depname;
  int  maxdep, ndep;
  int  *dep;

  sp = pkv_GetScratchMemTop ();
  maxdep = reader->maxdep;
  if ( maxdep > 0 )
    dep = pkv_GetScratchMemi ( maxdep );
  else
    dep = NULL;
  depname = reader->sc.nextsymbol;  /* it is supposed to be one of keywords */
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
                            /* now read the list of identifiers of */
                            /* objects with dependencies */
  ndep = 0;
  while ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
    if ( ndep >= maxdep )
      goto failure;
    if ( dep )
      dep[ndep] = reader->sc.nextinteger;
    ndep ++;
    reader->sc.GetNextSymbol ( &reader->sc );
    switch ( reader->sc.nextsymbol ) {
  case PKS_SYMB_COMMA:
      reader->sc.GetNextSymbol ( &reader->sc );
      break;
  case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      goto finish;
  default:
      goto failure;
    }
  }
finish:
  if ( reader->DepReader )
    reader->DepReader ( reader->userData, depname, ndep, dep );

  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*bsf_ReadDependencies*/

