
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017, 2018                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "pklinkage.h"
#include "bsfile.h"

#include "bsfprivate.h"

#define KWD(m) bsf_keyword[m]

static boolean bsf_WriteKLObject ( bsf_Writer *writer, pkkl_linkage *lkg, int i )
{
  int           sci; 
  pkkl_geom_obj *obj;

  obj = &lkg->obj[i];
  BSFwci
  sci = writer->current_indentation;
  writer->current_indentation += 2;
  writer->current_length += fprintf ( writer->output, "%s %d {",
            KWD(BSF_KEY_OBJECT), i );
  switch ( obj->type ) {
case PKKL_BSPLINE_CURVE:
    writer->current_length += fprintf ( writer->output, " %s",
              KWD(BSF_KEY_BSCURVE) );
    break;
case PKKL_BSPLINE_PATCH:
    writer->current_length += fprintf ( writer->output, " %s",
              KWD(BSF_KEY_BSPATCH) );
    break;
case PKKL_BSPLINE_MESH:
    writer->current_length += fprintf ( writer->output, " %s",
              KWD(BSF_KEY_BSMESH) );
    break;
default:
    if ( obj->name )
      writer->current_length += fprintf ( writer->output, " %s",
                KWD(BSF_KEY_NAME) );
    break;
  }
  if ( obj->name ) {
    writer->current_length += fprintf ( writer->output, " \"%s\"",
              obj->name );
    BSFdol
  }
  if ( obj->filename ) {
    writer->current_length += fprintf ( writer->output, " %s \"%s\" ",
              KWD(BSF_KEY_FILENAME), obj->filename );
    BSFdol
  }
  writer->current_length += fprintf ( writer->output, "%s %d ",
              KWD(BSF_KEY_DIM), obj->nvc );
  BSFdol
  writer->current_length += fprintf ( writer->output, "%s %d",
              KWD(BSF_KEY_NUM_VERTICES), obj->nvert );
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s ",
            KWD(BSF_KEY_ETRANS) );
  writer->current_indentation += 2;
  bsf_WriteTrans3d ( writer, &obj->Etr );
  writer->current_indentation -= 2;
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation = sci;
  return true;
} /*bsf_WriteKLObject*/

static boolean bsf_WriteKLObjectRef ( bsf_Writer *writer, pkkl_linkage *lkg, int i )
{
  pkkl_obj_ref *oref;
  int          j, sci;

  oref = &lkg->oref[i];
  sci = writer->current_indentation;
  BSFwci
  writer->current_indentation += 2;
  writer->current_length += fprintf ( writer->output, "%s %d %d {",
            KWD(BSF_KEY_OBJECT_REF), i, oref->on );
  if ( oref->vn ) {
    for ( j = 0; j < oref->nv; j++ ) {
      writer->current_length += fprintf ( writer->output, "%d",
                oref->vn[j] );
      if ( j < oref->nv-1 )
        writer->current_length += fprintf ( writer->output, "," );
      BSFdol
    }
  }
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation = sci;
  return true;
} /*bsf_WriteKLObjectRef*/

static boolean bsf_WriteKLLink ( bsf_Writer *writer, pkkl_linkage *lkg, int i )
{
  pkkl_link *lk;
  int       sci, rn;

  lk = &lkg->link[i];
  sci = writer->current_indentation;
  BSFwci
  writer->current_indentation += 2;
  writer->current_length += fprintf ( writer->output, "%s %d { ",
            KWD(BSF_KEY_LINK), i );
  if ( lk->name ) {
    writer->current_length += fprintf ( writer->output, "%s \"%s\"",
            KWD(BSF_KEY_NAME), lk->name );
    BSFdol
  }
  for ( rn = lk->fref; rn >= 0; rn = lkg->oref[rn].nextr ) {
    writer->current_length += fprintf ( writer->output, "%d", rn );
    if ( lkg->oref[rn].nextr >= 0 )
      writer->current_length += fprintf ( writer->output, "," );
    BSFdol
  }
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation = sci;
  return true;
} /*bsf_WriteKLLink*/

static boolean bsf_WriteJointArt ( bsf_Writer *writer, pkkl_linkage *lkg,
                                   pkkl_halfjoint *jt )
{
  switch ( jt->art ) {
case PKKL_ART_NONE:
    break;
case PKKL_ART_TRANS_X:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_TRANSLATION_X) );
    break;
case PKKL_ART_TRANS_Y:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_TRANSLATION_Y) );
    break;
case PKKL_ART_TRANS_Z:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_TRANSLATION_Z) );
    break;
case PKKL_ART_TRANS_V:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_TRANSLATION_V) );
    break;
case PKKL_ART_SCALE_X:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_SCALING_X) );
    break;
case PKKL_ART_SCALE_Y:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_SCALING_Y) );
    break;
case PKKL_ART_SCALE_Z:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_SCALING_Z) );
    break;
case PKKL_ART_SCALE_XYZ:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_SCALING_XYZ) );
    break;
case PKKL_ART_ROT_X:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_ROTATION_X) );
    break;
case PKKL_ART_ROT_Y:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_ROTATION_Y) );
    break;
case PKKL_ART_ROT_Z:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_ROTATION_Z) );
    break;
case PKKL_ART_ROT_V:
    writer->current_length += fprintf ( writer->output, "%s",
                KWD(BSF_KEY_ROTATION_V) );
    break;
default:
    return false;
  }
  writer->current_length += fprintf ( writer->output, " %d",
             jt->pnum );
  return true;
} /*bsf_WriteJointArt*/

static boolean bsf_WriteKLJoint ( bsf_Writer *writer, pkkl_linkage *lkg, int i )
{
  pkkl_halfjoint *jt;
  int            sci;

  jt = &lkg->hj[i];
  if ( jt->art != PKKL_ART_OTHERHALF ) {
    sci = writer->current_indentation;
    BSFwci
    writer->current_indentation += 2;
    writer->current_length += fprintf ( writer->output, "%s { ",
              KWD(BSF_KEY_JOINT) );
    writer->current_length += fprintf ( writer->output, "%d %d ",
              jt->l0, jt->l1 );
    bsf_WriteJointArt ( writer, lkg, jt );
    BSFeol BSFwci
    writer->current_length += fprintf ( writer->output, "%s ",
              KWD(BSF_KEY_FTRANS) );
    writer->current_indentation += 2;
    bsf_WriteTrans3d ( writer, &jt->Ftr );
    writer->current_indentation -= 2;
    BSFeol BSFwci
    writer->current_length += fprintf ( writer->output, "%s ",
              KWD(BSF_KEY_BTRANS) );
    writer->current_indentation += 2;
    bsf_WriteTrans3d ( writer, &jt->Btr );
    writer->current_indentation -= 2;
    writer->current_length += fprintf ( writer->output, "}" );
    BSFeol
    writer->current_indentation = sci;
  }
  return true;
} /*bsf_WriteKLJoint*/

static boolean bsf_WriteKLArtParam ( bsf_Writer *writer, pkkl_linkage *lkg )
{
  int i;

  writer->current_length += fprintf ( writer->output, "%s {",
                                KWD(BSF_KEY_ART_PARAM_RANGE) );
  writer->current_indentation += 2;
  for ( i = 0; i < lkg->nartpar; i++ ) {
    bsf_WriteDoubleNumber ( writer, lkg->artprange[2*i] );
    writer->current_length += fprintf ( writer->output, "," );
    bsf_WriteDoubleNumber ( writer, lkg->artprange[2*i+1] );
    if ( i < lkg->nartpar-1 ) {
      writer->current_length += fprintf ( writer->output, ";" );
      BSFdol  BSFwci
    }
    else
      writer->current_length += fprintf ( writer->output, "}" );
  }
  writer->current_indentation -= 2;
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s {",
                                KWD(BSF_KEY_ART_PARAM) );
  writer->current_indentation += 2;
  for ( i = 0; i < lkg->nartpar; i++ ) {
    bsf_WriteDoubleNumber ( writer, lkg->artp[i] );
    if ( i < lkg->nartpar-1 ) {
      writer->current_length += fprintf ( writer->output, ";" );
      BSFdol  BSFwci
    }
    else
       writer->current_length += fprintf ( writer->output, "}" );
  }
  writer->current_indentation -= 2;
  BSFeol  BSFwci
  return true;
} /*bsf_WriteKLArtParam*/

boolean bsf_WriteKLinkaged ( bsf_Writer *writer, pkkl_linkage *lkg,
                             const char *name )
{
  int     sci;
  boolean result;
  int     i;

  result = true;
  sci = writer->current_indentation;
  BSFwci
  fprintf ( writer->output, "%s\n", "%Kinematic linkage" );
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s {",
            KWD(BSF_KEY_KLINKAGE) );
  writer->current_indentation += 2;
  BSFeol
  if ( name && *name ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s \"%s\"",
                KWD(BSF_KEY_NAME), name ); 
    BSFeol
  }
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            KWD(BSF_KEY_NUM_OBJECTS), lkg->nobj );
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            KWD(BSF_KEY_NUM_OBJ_REF), lkg->norefs );
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            KWD(BSF_KEY_NUM_LINKS), lkg->nlinks );
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            KWD(BSF_KEY_NUM_JOINTS), lkg->nhj/2 );
  BSFeol  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d",
            KWD(BSF_KEY_NUM_ART_PAR), lkg->nartpar );
  BSFeol
  for ( i = 0; i < lkg->nobj; i++ )
    result &= bsf_WriteKLObject ( writer, lkg, i );
  for ( i = 0; i < lkg->nlinks; i++ )
    result &= bsf_WriteKLLink ( writer, lkg, i );
  for ( i = 0; i < lkg->norefs; i++ )
    result &= bsf_WriteKLObjectRef ( writer, lkg, i );
  for ( i = 0; i < lkg->nhj; i++ )
    result &= bsf_WriteKLJoint ( writer, lkg, i );
  if ( lkg->nartpar > 0 ) {
    BSFwci
    bsf_WriteKLArtParam ( writer, lkg );
  }
  writer->current_indentation = sci;
  BSFwci
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  return result;
} /*bsf_WriteKLinkaged*/

