
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "bsmesh.h"

#include "bsmprivate.h"

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _bsm_VertOk ( const BSMvertex *mv, const int *mvhei,
                             const BSMhalfedge *mhe, int v )
{
  int d;

  if ( (d = mv[v].degree) < 1 )
    return false;
  if ( d < 4 )
    return mhe[mvhei[mv[v].firsthalfedge+d-1]].otherhalf < 0;
  return true;
} /*_bsm_VertOk*/

boolean bsm_RemoveEdgeNum ( int inv, const BSMvertex *imv, const int *imvhei,
                            int inhe, BSMhalfedge *imhe,
                            int infac, const BSMfacet *imfac, const int *imfhei,
                            int he,
                            int *onv, int *onhe, int *onfac )
{
  int he1, h, f0, f1, i, fhe, deg;

  if ( he < 0 || he >= inhe )
    return false;
  if ( (he1 = imhe[he].otherhalf) < 0 )  /* only an inner edge */
    return false;
                                 /* with inner vertices */
  if ( !_bsm_VertOk ( imv, imvhei, imhe, imhe[he].v0 ) ||
       !_bsm_VertOk ( imv, imvhei, imhe, imhe[he].v1 ) )
    return false;
  f0 = imhe[he].facetnum;
  f1 = imhe[he1].facetnum;
  fhe = imfac[f1].firsthalfedge;
  deg = imfac[f1].degree;
  for ( i = 0; i < deg; i++ )
    if ( (h = imfhei[fhe+i]) != he1 ) {
      if ( h >= 0 && imhe[h].facetnum == f0 )
        return false;
    }
  *onv = inv;
  *onhe = inhe-2;
  *onfac = infac-1;
  return true;
} /*bsm_RemoveEdgeNum*/

boolean bsm_RemoveEdged ( int spdimen, int inv,
                          const BSMvertex *imv, const int *imvhei,
                          double *iptc, int inhe, const BSMhalfedge *imhe,
                          int infac, const BSMfacet *imfac, const int *imfhei,
                          int he,
                          int *onv, BSMvertex *omv, int *omvhei,
                          double *optc, int *onhe, BSMhalfedge *omhe,
                          int *onfac, BSMfacet *omfac, int *omfhei )
{
  int f0, f1, v0, v1, he0, he1, h;
  int v, i, j, f, g, deg, dd, ifhe, ofhe;

  if ( he < 0 || he >= inhe )
    return false;
  if ( imhe[he].otherhalf < 0 )  /* only an inner edge */
    return false;
  if ( he < imhe[he].otherhalf ) {
    he0 = he;
    he1 = imhe[he].otherhalf;
  }
  else {
    he0 = imhe[he].otherhalf;
    he1 = he;
  }
  v0 = imhe[he0].v0;
  v1 = imhe[he0].v1;
  f0 = imhe[he0].facetnum;
  f1 = imhe[he1].facetnum;
  if ( !_bsm_VertOk ( imv, imvhei, imhe, v0 ) ||
       !_bsm_VertOk ( imv, imvhei, imhe, v1 ) )
    return false;

  ifhe = imfac[f1].firsthalfedge;
  deg = imfac[f1].degree;
  for ( i = 0; i < deg; i++ )
    if ( (h = imfhei[ifhe+i]) != he1 ) {
      if ( h >= 0 && imhe[h].facetnum == f0 )
        return false;
    }
        /* vertex positions do not change */
  memcpy ( optc, iptc, spdimen*inv*sizeof(double) );
  memcpy ( omv, imv, inv*sizeof(BSMvertex) );
  omv[v0].degree --;
  omv[v1].degree --;
        /* copy the original halfedge numbers, to be modified */
          /* vertices first */
  for ( v = v0+1; v < inv; v++ )
    omv[v].firsthalfedge --;
  for ( v = v1+1; v < inv; v++ )
    omv[v].firsthalfedge --;
  for ( v = 0; v < inv; v++ ) {
    deg = imv[v].degree;
    ifhe = imv[v].firsthalfedge;
    ofhe = omv[v].firsthalfedge;
    if ( v == v0 ) {
      for ( i = 0, j = ofhe;  i < deg;  i++ )
        if ( imvhei[ifhe+i] != he0 )
          omvhei[j++] = imvhei[ifhe+i];
    }
    else if ( v == v1 ) {
      for ( i = 0, j = ofhe;  i < deg;  i++ )
        if ( imvhei[ifhe+i] != he1 )
          omvhei[j++] = imvhei[ifhe+i];
    }
    else
      memcpy ( &omvhei[ofhe], &imvhei[ifhe], deg*sizeof(int) );
  }
          /* then halfedges */
  for ( i = j = 0;  i < inhe;  i++ )
    if ( i != he0 && i != he1 ) {
      omhe[j] = imhe[i];
      if ( imhe[i].otherhalf > he0 )
        omhe[j].otherhalf --;
      if ( imhe[i].otherhalf > he1 )
        omhe[j].otherhalf --;
      j ++;
    }
        /* copy the facets */
  for ( f = g = ofhe = 0;  f < infac;  f++ ) {
    ifhe = imfac[f].firsthalfedge;
    deg = imfac[f].degree;
    if ( f == f0 ) {  /* join facet f1 to f0 */
      omfac[g].firsthalfedge = ofhe;
      dd = deg + imfac[f1].degree - 2;
      omfac[g].degree = dd;
      for ( i = 0; i < deg && imfhei[ifhe+i] != he0; i++ )
        ;
      if ( i == deg )
        return false;  /* an invalid mesh */
      for ( j = 0, i = (i+1) % deg;  j < deg-1;  j++, i = (i+1) % deg )
        omfhei[ofhe+j] = imfhei[ifhe+i];
      ofhe += deg-1;
      ifhe = imfac[f1].firsthalfedge;
      deg = imfac[f1].degree;
      for ( i = 0; i < deg && imfhei[ifhe+i] != he1; i++ )
        ;
      if ( i == deg )
        return false;  /* an invalid mesh */
      for ( j = 0, i = (i+1) % deg;  j < deg-1;  j++, i = (i+1) % deg )
        omfhei[ofhe+j] = imfhei[ifhe+i];
      ofhe += deg-1;
      g ++;
    }
    else if ( f != f1 ) {
      omfac[g].firsthalfedge = ofhe;
      omfac[g].degree = deg;
      memcpy ( &omfhei[ofhe], &imfhei[ifhe], deg*sizeof(int) );
      ofhe += deg;
      g ++;
    }
  }
        /* renumber the halfedges */
  for ( i = 0; i < inhe-2; i++ ) {
    if ( omvhei[i] > he1 )
      omvhei[i] --;
    if ( omvhei[i] > he0 )
      omvhei[i] --;
  }
  for ( i = 0; i < inhe-2; i++ ) {
    if ( omfhei[i] > he1 )
      omfhei[i] --;
    if ( omfhei[i] > he0 )
      omfhei[i] --;
  }
        /* assign the halfedge facets */
  for ( f = 0; f < infac-1; f++ ) {
    deg = omfac[f].degree;
    ofhe = omfac[f].firsthalfedge;
    for ( j = 0; j < deg; j++ )
      omhe[omfhei[ofhe+j]].facetnum = f;
  }

  *onv = inv;
  *onhe = inhe-2;
  *onfac = infac-1;
  return true;
} /*bsm_RemoveEdged*/

