
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "multibs.h"

boolean mbs_BezierStandardC2Rd ( int degree, point3d *cpoints )
{
  int    i;
  double w0, wn, r0, rn, r, s;

  if ( degree < 0 )
    return false;
  w0 = cpoints[0].z;
  wn = cpoints[degree].z;
  if ( pkv_signd ( w0 )*pkv_signd ( wn ) <= 0 )
    return false;
  SetPoint3d ( cpoints,
               cpoints[0].x/cpoints[0].z,
               cpoints[0].y/cpoints[0].z, 1.0 );
  if ( degree > 0 ) {
    if ( w0 < 0.0 ) {
      for ( i = 0; i <= degree; i++ ) {
        cpoints[i].x = -cpoints[i].x;
        cpoints[i].y = -cpoints[i].y;
        cpoints[i].z = -cpoints[i].z;
      }
      w0 = cpoints[0].z;
      wn = cpoints[degree].z;
    }
    SetPoint3d ( &cpoints[degree],
                 cpoints[degree].x/cpoints[degree].z,
                 cpoints[degree].y/cpoints[degree].z, 1.0 );
    if ( degree > 1 ) {
      s = 1.0/(double)degree;
      r0 = pow ( w0, s );
      rn = pow ( wn, s );
      s = r0/rn;
      r = s/w0;
      for ( i = 1; i < degree; i++ ) {
        MultVector3d ( r, &cpoints[i], &cpoints[i] );
        r *= s;
      }
    }
  }
  return true;
} /*mbs_BezierStandardC2Rd*/

