
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017, 2018                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "pklinkage.h"

/* ///////////////////////////////////////////////////////////////////////// */
boolean pkkl_LinkageValid ( pkkl_linkage *linkage )
{
  int            i, j, k, l;
  pkkl_geom_obj  *obj, *ob;
  pkkl_obj_ref   *oref;
  pkkl_link      *link;
  pkkl_halfjoint *hj, *hj0, *hj1;

  if ( linkage->nobj < 0 || linkage->nobj > linkage->maxobj )
    return false;
  if ( linkage->nlinks < 0 || linkage->nlinks > linkage->maxlinks )
    return false;
  if ( linkage->norefs < 0 || linkage->norefs > linkage->maxorefs )
    return false;
  if ( linkage->nhj < 0 || linkage->nhj > linkage->maxnhj )
    return false;
  if ( linkage->nartpar < 0 || linkage->nartpar > linkage->maxartpar )
    return false;
  if ( linkage->current_root < 0 ||
       (linkage->nlinks > 0 && linkage->current_root >= linkage->nlinks) )
    return false;
  obj = linkage->obj;
  oref = linkage->oref;
  link = linkage->link;
  hj = linkage->hj;
  if ( !obj || !oref || !link || !hj || !linkage->artp || !linkage->prevartp )
    return false;
        /* check the geometric objects */
  for ( i = 0; i < linkage->nobj; i++ ) {
    if ( (obj[i].nvc != 3 && obj[i].nvc != 4) ||
         obj[i].nvert < 1 /*|| !obj[i].vert || !obj[i].tvert*/ )
      return false;
  }
        /* check the object references */
  for ( i = 0; i < linkage->norefs; i++ ) {
    if ( oref[i].on < 0 || oref[i].on >= linkage->nobj )
      return false;
    if ( oref[i].nextr < -1 || oref[i].nextr >= linkage->norefs )
      return false;
    ob = &obj[oref[i].on];
    if ( oref[i].nv < 1 || oref[i].nv > ob->nvert )
      return false;
    if ( oref[i].vn ) {
      for ( k = 0; k < oref[i].nv; k++ )
        if ( oref[i].vn[k] < 0 || oref[i].vn[k] >= ob->nvert )
          return false;
    }
  }
        /* check the links */
  k = l = 0;
  for ( i = 0; i < linkage->nlinks; i++ ) {
    if ( link[i].fref < -1 || link[i].fref >= linkage->norefs )
      return false;
    for ( j = link[i].fref; j >= 0 && k <= linkage->norefs; j = oref[j].nextr )
      k ++;
    for ( j = link[i].fhj; j >= 0 && l <= linkage->nhj; j = hj[j].nexthj ) {
      if ( hj[j].l0 != i )
        return false;
      l ++;
    }
  }
  if ( k != linkage->norefs || l != linkage->nhj )
    return false;
        /* check the halfjoints */
  for ( i = 0; i < linkage->nhj; i++ ) {
    hj0 = &hj[i];
    if ( hj0->l0 < 0 || hj0->l0 >= linkage->nlinks ||
         hj0->l1 < 0 || hj0->l1 >= linkage->nlinks ||
         hj0->l0 == hj0->l1 )
      return false;
    k = hj0->otherhalf;
    if ( k < 0 || k >= linkage->nhj )
      return false;
    hj1 = &hj[k];
    if ( hj1->otherhalf != i || hj1->l0 != hj0->l1 || hj1->l1 != hj0->l0 )
      return false;
    if ( hj0->pnum < -1 || hj0->pnum >= linkage->maxartpar )
      return false;
    if ( hj0->art < PKKL_ART_OTHERHALF || hj0->art > PKKL_ART_AFFINE )
      return false;
  }

  return true;
} /*pkkl_LinkageValid*/

