
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "pklinkage.h"

/* ///////////////////////////////////////////////////////////////////////// */
void pkkl_SetArtParam ( pkkl_linkage *linkage, int pno, int nump, double *par )
{
  if ( nump > 0 && pno >= 0 && pno+nump <= linkage->maxartpar ) {
    memcpy ( &linkage->prevartp[pno], &linkage->artp[pno], nump*sizeof(double) );
    memcpy ( &linkage->artp[pno], par, nump*sizeof(double) );
    if ( pno+nump > linkage->nartpar )
      linkage->nartpar = pno+nump;
  }
} /*pkkl_SetArtParam*/

static boolean _pkkl_changed ( double *a, double *b, int i, int n )
{
  return memcmp ( &a[i], &b[i], n*sizeof(double) ) != 0;
} /*_pkkl_changed*/

static void _pkkl_UpdateArtTr ( pkkl_linkage *linkage, int jn )
{
  pkkl_halfjoint *hj, *hj1;
  int            pnum;
  double         *artp, *prevartp;
  quaterniond    q;

  artp = linkage->artp;
  prevartp = linkage->prevartp;
  hj = &linkage->hj[jn];
  hj1 = &linkage->hj[hj->otherhalf];
  if ( hj->art == PKKL_ART_OTHERHALF ) {
    hj = hj1;
    hj1 = &linkage->hj[hj->otherhalf];
    if ( hj1->art == PKKL_ART_OTHERHALF )
      return;
  }
  pnum = hj->pnum;
  switch ( hj->art ) {
case PKKL_ART_NONE:
    break;

case PKKL_ART_TRANS_X:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Shiftd ( &hj->Atr, artp[pnum], 0.0, 0.0 );
      ShiftTrans3d ( &hj1->Atr, -artp[pnum], 0.0, 0.0 );
    }
    break;

case PKKL_ART_TRANS_Y:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Shiftd ( &hj->Atr, 0.0, artp[pnum], 0.0 );
      ShiftTrans3d ( &hj1->Atr, 0.0, -artp[pnum], 0.0 );
    }
    break;

case PKKL_ART_TRANS_Z:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Shiftd ( &hj->Atr, 0.0, 0.0, artp[pnum] );
      ShiftTrans3d ( &hj1->Atr, 0.0, 0.0, -artp[pnum] );
    }
    break;

case PKKL_ART_TRANS_V:
    if ( _pkkl_changed ( artp, prevartp, pnum, 3 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Shiftd ( &hj->Atr, artp[pnum], artp[pnum+1], artp[pnum+2] );
      ShiftTrans3d ( &hj1->Atr, -artp[pnum], -artp[pnum+1], -artp[pnum+2] );
    }
    break;

case PKKL_ART_SCALE_X:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Scaled ( &hj->Atr, artp[pnum], 1.0, 1.0 );
      ScaleTrans3d ( &hj1->Atr, 1.0/artp[pnum],1.0, 1.0 );
    }
    break;

case PKKL_ART_SCALE_Y:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Scaled ( &hj->Atr, 1.0, artp[pnum], 1.0 );
      ScaleTrans3d ( &hj1->Atr, 1.0, 1.0/artp[pnum], 1.0 );
    }
    break;

case PKKL_ART_SCALE_Z:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Scaled ( &hj->Atr, 1.0, 1.0, artp[pnum] );
      ScaleTrans3d ( &hj1->Atr, 1.0, 1.0, 1.0/artp[pnum] );
    }
    break;

case PKKL_ART_SCALE_XYZ:
    if ( _pkkl_changed ( artp, prevartp, pnum, 3 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3Scaled ( &hj->Atr, artp[pnum], artp[pnum+1], artp[pnum+2] );
      ScaleTrans3d ( &hj1->Atr, 1.0/artp[pnum], 1.0/artp[pnum+1], 1.0/artp[pnum+2] );
    }
    break;

case PKKL_ART_ROT_X:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3RotXd ( &hj->Atr, artp[pnum] );
      RotXTrans3d ( &hj1->Atr, -artp[pnum] );
    }
    break;

case PKKL_ART_ROT_Y:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3RotYd ( &hj->Atr, artp[pnum] );
      RotYTrans3d ( &hj1->Atr, -artp[pnum] );
    }
    break;

case PKKL_ART_ROT_Z:
    if ( _pkkl_changed ( artp, prevartp, pnum, 1 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3RotZd ( &hj->Atr, artp[pnum] );
      RotZTrans3d ( &hj1->Atr, -artp[pnum] );
    }
    break;

case PKKL_ART_ROT_V:
    if ( _pkkl_changed ( artp, prevartp, pnum, 4 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3RotVd ( &hj->Atr, (vector3d*)&artp[pnum], artp[pnum+3] );
      RotVTrans3d ( &hj1->Atr, (vector3d*)&artp[pnum], -artp[pnum+3] );
    }
    break;

case PKKL_ART_ROT_Q:
    if ( _pkkl_changed ( artp, prevartp, pnum, 4 ) ) {
      IdentTrans3d ( &hj->Atr );
      IdentTrans3d ( &hj1->Atr );
      Trans3QuaternionRotd ( &hj->Atr, (quaterniond*)&artp[pnum] );
      memcpy ( &q, &artp[pnum], 3*sizeof(double) );
      q.w = -artp[pnum+3];
      QuaternionRotTrans3d ( &hj1->Atr, &q );
    }
    break;

case PKKL_ART_AFFINE:
    if ( _pkkl_changed ( artp, prevartp, pnum, 12 ) ) {
        /* detsgn is now ignored, to be fixed */
      memcpy ( &hj->Atr, &artp[pnum], 12*sizeof(double) );
      memcpy ( &hj1->Atr, &artp[pnum], 12*sizeof(double) );
      InvertTrans3d ( &hj1->Atr );
    }
    break;

default:
    IdentTrans3d ( &hj->Atr );
    IdentTrans3d ( &hj1->Atr );
    break;
  }
} /*_pkkl_UpdateArtTr*/

static void _pkkl_rArticulate ( pkkl_linkage *linkage, int lkn, trans3d *tr )
{
  pkkl_link      *lk;
  trans3d        t0, t1;
  int            r, on, on1, j, l1;
  pkkl_geom_obj  *obj;
  pkkl_obj_ref   *oref;
  pkkl_halfjoint *hj;

  obj = linkage->obj;
  oref = linkage->oref;
  hj = linkage->hj;
  lk = &linkage->link[lkn];
  if ( !lk->tag ) {
    lk->tag = 1;
    CompTrans3d ( &t0, tr, &lk->Mtr );
        /* pass through the list of object references and transform */
        /* the object vertices */
    for ( r = lk->fref, on1 = -1;  r >= 0;  r = oref[r].nextr ) {
      on = oref[r].on;
      if ( obj[on].transform ) {
        if ( on != on1 ) {
          CompTrans3d ( &t1, &t0, &obj[on].Etr );
          on1 = on;
        }
        obj[on].transform ( &obj[on], &t1, oref[r].nv, oref[r].vn );
      }
    }
        /* now traverse the list of halfjoints originating at */
        /* the current link to (DFS) search the linkage graph */
    for ( j = lk->fhj; j >= 0; j = hj[j].nexthj ) {
      l1 = hj[j].l1;
      if ( !linkage->link[l1].tag ) {
        CompTrans3d ( &t0, tr, &hj[j].Ftr );
        CompTrans3d ( &t1, &t0, &hj[j].Atr );
        CompTrans3d ( &t0, &t1, &hj[j].Btr );
        _pkkl_rArticulate ( linkage, l1, &t0 );
      }
    }
  }
} /*_pkkl_rArticulate*/

void pkkl_Articulate ( pkkl_linkage *linkage )
{
  int i;

  for ( i = 0; i < linkage->nlinks; i++ )
    linkage->link[i].tag = 0;
  for ( i = 0; i < linkage->nhj; i++ )
    _pkkl_UpdateArtTr ( linkage, i );
  _pkkl_rArticulate ( linkage, linkage->current_root, &linkage->current_root_tr );
  for ( i = 0; i < linkage->nobj; i++ )
    linkage->obj[i].postprocess ( &linkage->obj[i] );
  memcpy ( linkage->prevartp, linkage->artp, linkage->maxartpar*sizeof(double) );
} /*pkkl_Articulate*/

void pkkl_Redraw ( pkkl_linkage *linkage, void *appdata )
{
  int i;

  for ( i = 0; i < linkage->nobj; i++ )
    if ( linkage->obj[i].redraw )
      linkage->obj[i].redraw ( &linkage->obj[i], linkage->usrdata, appdata );
} /*pkkl_Redraw*/

