
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"

boolean pkn_multiSolveDLSQd ( int nrows, int ncols, const double *a,
                              int spdimen, int bpitch, double *b,
                              int xpitch, const double *xd, double *x )
{
  void   *sp;
  double *at, *aa, *z, *w;

  sp = pkv_GetScratchMemTop ();
  if ( nrows >= ncols )
    goto failure;
  at = pkv_GetScratchMemd ( nrows*(ncols+2)+(ncols+nrows)*spdimen );
  if ( !at )
    goto failure;
  aa = &at[nrows*ncols];
  z = &aa[2*nrows];
  w = &z[nrows*spdimen];
  pkv_TransposeMatrixd ( nrows, ncols, ncols, a, nrows, at );
  pkn_QRDecomposeMatrixd ( ncols, nrows, at, aa );
  pkn_multiMultInvTrUTVectord ( nrows, at, spdimen, bpitch, b, spdimen, z );
  if ( xd ) {
    pkv_Selectd ( ncols, spdimen, xpitch, spdimen, xd, w );
    pkn_multiReflectVectord ( ncols, nrows, at, aa, spdimen, spdimen, w );
    pkn_SubtractMatrixd ( 1, nrows*spdimen, 0, z, 0, w, 0, z );
  }
  memset ( &z[nrows*spdimen], 0, (ncols-nrows)*spdimen*sizeof(double) );
  pkn_multiInvReflectVectord ( ncols, nrows, at, aa, spdimen, spdimen, z );
  pkn_AddMatrixd ( ncols, spdimen, spdimen, z, xpitch, xd, xpitch, x );

  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*pkn_multiSolveDLSQd*/

