
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

    /* compiled batch statements - each of them follows PKS_SYMB_BATCH_CMD */
#define PKS_BATCHCMD_JUMP      1
#define PKS_BATCHCMD_GOTO      2
#define PKS_BATCHCMD_LABEL     3
#define PKS_BATCHCMD_CASES     4
#define PKS_BATCHCMD_CONTINUE  5
#define PKS_BATCHCMD_IF        6
#define PKS_BATCHCMD_WHILE     7
#define PKS_BATCHCMD_FOR       8
#define PKS_BATCHCMD_SWITCH    9
#define PKS_BATCHCMD_INCLUDE  10
#define PKS_BATCHCMD_ENDINCL  11

/* ///////////////////////////////////////////////////////////////////////// */
typedef struct {
          short int token, key;
        } keystatement;

typedef struct {
          short int token, key;
          int       dest;
        } jumpstatement;

typedef struct {
          short int token, key;
          int       ifptr, elseptr;
        } ifstatement;

typedef struct {   
          short int token, key;
          int       loopptr, endptr;
        } whileloop;

typedef struct {
          short int token, key;
          int       initial, final, incr, loopptr, endptr;
          char      level;    /* for loop nesting level */
        } forloop;

typedef struct {
          short int token, key;
          int       cases;
        } switchstatement;

typedef struct {
          short int token, key;
          short int ncases;
          int       defaultloc;
        } switchcases;

typedef struct {
          short int token, key;
        } include_filename;

typedef struct {
          short int token, key;
          int       namepos;
        } include_end;

typedef struct {
          short int token, level;
        } forcounter;

    /* auxiliary data used during the compilation */
typedef struct forcntname {
          char              *name;
          struct forcntname *next;
        } forcntname;

typedef struct labeldesc {
          struct labeldesc *next;     /* next label description */
          int              loc;       /* labelled spot location */
          int              caseval;
          char             name[1];   /* label name */
        } labeldesc;

typedef struct {
          struct labeldesc *gotos;
          struct labeldesc *labels;
          struct labeldesc *breaks;
          struct labeldesc *continues;
          struct labeldesc *cases;
          int              defaultloc;
        } block_labels;

typedef struct {
          boolean    last_semicolon;  /* how was the last token sequence ended */
          short int  fornesting, maxfornesting;
          forcntname *firstcnt;       /* names of for loop control variables */
        } auxcomp;

/* ///////////////////////////////////////////////////////////////////////// */
int _pksbatch_GetNextSymbolFSC ( pks_scanner *sc );
int _pksbatch_GetNextSymbolCSC ( pks_scanner *sc );

