
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

/* ///////////////////////////////////////////////////////////////////////// */
/* representation of tokens in parsed macros */
typedef struct {  /* PKS_SYMB_IDENT, PKS_SYMB_STRING */
    short symbol;  /* the first two fields must be like these of pks_token_any */
    short length;
    int   intvalue;
    short id_space;
    char  name[1];
  } pks_token_name;

typedef struct {  /* PKS_SYMB_INTEGER, PKS_SYMB_CHAR, PKS_SYMB_MACROPAR */
    short symbol;
    short length;
    int   intvalue;
  } pks_token_int; 

typedef struct {  /* PKS_SYMB_FLOAT */
    short symbol;
    short length;
    double floatvalue;
  } pks_token_float;  

typedef struct {  /* none of the above */
    short symbol;
    short length;
  } pks_token_any;

typedef union pks_token {
    pks_token_any   any; 
    pks_token_name  name;
    pks_token_int   integer;
    pks_token_float real;   
  } pks_token;

/* this data structure is immediately followed by the macro name, */
/* then by parameter names and the sequence of tokens */
typedef struct pks_macro {
    AVLtnode         node;
    struct pks_macro *previous; /* link to stack of macros */
    int              nsymb;     /* next symbol to return */  
    char             nparams;
    int              params;    /* offset in bytes to the first parameter macro */
    int              tokens;    /* offset in bytes to the first token */
  } pks_macro;

typedef struct {
    AVLtnode node;
    short    number;   /* which formal parameter is this ? */
    char     name[1];  /* parameter name */
  } pks_macropname;

/* ///////////////////////////////////////////////////////////////////////// */
extern boolean _pks_SkipComment ( pks_scanner *sc, const char *bc, const char *ec );
extern boolean _pks_AddTheMacroToken ( pks_scanner *psc, char *mbuf, int *bufp );

