
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkgeom.h"
#include "multibs.h"

#include "xgedit.h"
#include "xgeprivate.h"

boolean _xge_InitInit ( int argc, char *argv[] )
{
  _xge_argc = argc;
  _xge_argv = argv;

  xge_p_name = argv[0];
  xgedisplay = XOpenDisplay ( "" );
  if ( !xgedisplay ) {
    printf ( "Error: Could not connect to the X server\n" );
    return false;
  }
  xgescreen = DefaultScreen ( xgedisplay );
  xgeroot = RootWindow ( xgedisplay, xgescreen );
  return true;
} /*_xge_InitInit*/

void _xge_InitWinHints ( void )
{
  xgehints.flags = PPosition | PSize | PMinSize;
  xgehints.height = xge_HEIGHT;
  xgehints.width = xge_WIDTH;
  xgehints.x = 0;
  xgehints.y = 100;
  xgehints.min_height = xge_HEIGHT / 2;
  xgehints.min_width = xge_WIDTH / 2;
} /*_xge_InitWinHints*/

void _xge_InitGCandCursors ( void )
{
  xgegc = XCreateGC ( xgedisplay, xgewindow, 0, 0 );
  xge_NewCursor ( XC_crosshair );  /* 0 four windows widget */
  xge_NewCursor ( XC_hand1 );      /* 1 list box */   
  xge_NewCursor ( XC_pencil );     /* 2 text editor */
  xge_NewCursor ( XC_fleur );      /* 3 */
  xge_NewCursor ( XC_draft_large /* XC_arrow */ );      /* 4 */
  xge_NewCursor ( XC_watch );      /* 5 */
  xge_NewCursor ( XC_circle );     /* 6 */
  xge_NewCursor ( XC_left_ptr );   /* 7 default arrow cursor */
  xge_NewCursor ( -1 );            /* 8 invisible (pixmap) cursor */
  xge_SetCurrentWindowCursor ( xgeCURSOR_DEFAULT );
} /*_xge_InitGCandCursors*/

void _xge_FinishInit ( void )
{
  xge_InitRectAllocation ();
  xge_errmsg_edr = xge_NewEmptyWidget ( -1, NULL, -1, 1, 1, -1, -1 );
  _xge_background_widget = xge_NewWidget ( -1, NULL, -1, 1, 1, -1, -1, NULL, NULL,
                                           _xge_background_msg, xge_DrawEmpty );  
  _xge_special_widget = xge_NewEmptyWidget ( -1, NULL, -1, 1, 1, -1, -1 );
  xge_null_widget  = xge_NewEmptyWidget ( -1, NULL, -1, 1, 1, -1, -1 );   
} /*_xge_FinishInit*/

