
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "xgedit.h"
#include "xgeprivate.h"


XImage      *_xge_canvas     = NULL;
static char *_xge_canvasdata = NULL;

boolean _xge_InitCanvas ( void )
{
  int nplanes;

  if ( !_xge_canvas ) {
    nplanes = XDisplayPlanes ( xgedisplay, xgescreen );
    _xge_canvas = XCreateImage ( xgedisplay, xgevisual, nplanes, ZPixmap,
                           0, NULL, XGE_CANVAS_WIDTH, XGE_CANVAS_HEIGHT, 8, 0 );
    if ( !_xge_canvas )
      return false;
    _xge_canvasdata =
        (char*)malloc ( XGE_CANVAS_HEIGHT*_xge_canvas->bytes_per_line );
    if ( !_xge_canvasdata ) {
      XDestroyImage ( _xge_canvas );
      _xge_canvas = NULL;
      return false;
    }
    memset ( _xge_canvasdata, 0, XGE_CANVAS_HEIGHT*_xge_canvas->bytes_per_line );
    _xge_canvas->data = _xge_canvasdata;
  }
  return true;
} /*_xge_InitCanvas*/

static xgecolour_int _GetAAPixel ( short width, byte *data, short x, short y )
{
  const static unsigned int fc[7] = {4,22,60,84,60,22,4};
  int                       i, j, X, Y, Z;
  unsigned int              r[7], g[7], b[7];

  for ( i = 0; i < 7; i++ ) {
    X = 3*(3*x+i);
    r[i] = g[i] = b[i] = 0;
    for ( j = 0; j < 7; j++ ) {
      Y = 3*y+j;
      Z = 3*Y*width+X;
      r[i] += fc[j]*data[Z];
      g[i] += fc[j]*data[Z+1];
      b[i] += fc[j]*data[Z+2];
    }
  }
  r[0] *= fc[0];  g[0] *= fc[0];  b[0] *= fc[0];
  for ( i = 1; i < 7; i++ ) {
    r[0] += fc[i]*r[i];
    g[0] += fc[i]*g[i];
    b[0] += fc[i]*b[i];
  }
  return xge_PixelColour ( r[0] >> 16, g[0] >> 16, b[0] >> 16 );
} /*_GetAAPixel*/

boolean _xge_DrawOnCanvasAA ( short w, short h, short xc, short yc,
                              short imw, short imh, byte *data )
{
  int x, y;

  if ( imw < 3*w+4 || imh < 2*h+4 || xc < 0 || yc < 0 ||
       xc+w > XGE_CANVAS_WIDTH || yc+h > XGE_CANVAS_HEIGHT )
    return false;
  for ( y = 0;  y < h;  y++ )
    for ( x = 0;  x < w;  x++ )
      XPutPixel ( _xge_canvas, xc+x, yc+y, _GetAAPixel ( imw, data, x, y ) );
  return true;
} /*_xge_DrawOnCanvasAA*/

