
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"

#include "xgedit.h"
#include "xgeprivate.h"

#define xgeid_FILEMENU_OPEN_DIRTEXT   0
#define xgeid_FILEMENU_OPEN_SW_HIDDEN 1
#define xgeid_FILEMENU_OPEN_FILE      2
#define xgeid_FILEMENU_OPEN_CANCEL    3
#define xgeid_FILEMENU_OPEN_DIRLIST   4
#define xgeid_FILEMENU_OPEN_FILELIST  5

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _xge_OpenFileProc ( xge_widget *er,
                                   int msg, int key, short x, short y )
{
  xge_widget   *up;
  xge_filemenu *fm;

  up = er->up;
  fm = (xge_filemenu*)up->data2;
  switch ( msg ) {
case xgemsg_BUTTON_COMMAND:
    switch ( er->id ) {
  case xgeid_FILEMENU_OPEN_FILE:
      xge_callback ( up, xgemsg_FILEMENU_OPENFILE, 0, 0, 0 );
      return true;
  case xgeid_FILEMENU_OPEN_CANCEL:
      xge_callback ( up, xgemsg_FILEMENU_CANCEL, 0, 0, 0 );
      xge_RemovePopup ( true );
      return true;
  default:
      return false;
    }

case xgemsg_SWITCH_COMMAND:
    switch ( er->id ) {
  case xgeid_FILEMENU_OPEN_SW_HIDDEN:
      return true;
  default:
      return false;
    }

case xgemsg_TEXT_WIDGET_CLICK:
    switch ( er->id ) {
  case xgeid_FILEMENU_OPEN_DIRTEXT:
      xge_ChangeDirAlt ( fm, x, true );
      xge_SetClipping ( fm->er1 );
      fm->er1->redraw ( fm->er1, true );
      return true;
  default:
      return false;
    }

case xgemsg_LISTBOX_ITEM_PICK:
    switch ( er->id ) {
  case xgeid_FILEMENU_OPEN_DIRLIST:
      xge_ChangeDir ( fm, true );
      xge_SetClipping ( fm->er1 );
      fm->er1->redraw ( fm->er1, true );
      return true;
  case xgeid_FILEMENU_OPEN_FILELIST:
      if ( xge_GetCurrentListBoxString ( &fm->filelist1, fm->filename ) ) {
        fm->lastfileitem1 = fm->filelist1.currentitem;
        xge_callback ( up, xgemsg_FILEMENU_OPENFILE, 0, 0, 0 );
            /* the popup menu ought to be removed by the application */
            /* in case it has to stay because of an error or something */
      }
      return true;
  default:
      return false;
    }

default:
    return false;
  }
} /*_xge_OpenFileProc*/

boolean xge_AddOpenFilePopup ( xge_filemenu *fm )
{
  xge_SetCurrentDir ( fm, true );
  xge_AddPopup ( fm->er1, true, true );
  return true;
} /*xge_AddOpenFilePopup*/

xge_widget *xge_InitOpenFileMenu ( char window_num, int id,
                                   short w, short h, short x, short y,
                                   xge_filemenu *fm, xge_widget *optw,
                                   char *fullpath, char *shortpath,
                                   char *prevpath, char *filter,
                                   char *filename )
{
  xge_widget *mw, *menu;
  int        hl;

  if ( w < 400 ) w = 400;
  if ( h < 232 ) h = 232;
  fm->fullpath      = fullpath;
  fm->shortpath     = shortpath;
  fm->prevpath      = prevpath;
  fm->filter        = filter;
  fm->filename      = filename;
  fm->showhidden    = false;
  fm->lastfileitem1 = 0;
  mw = xge_NewTextWidget ( window_num, optw, xgeid_FILEMENU_OPEN_DIRTEXT,
                           380, 16, x+10, y+10, shortpath );
  mw = xge_NewSwitch ( window_num, mw, xgeid_FILEMENU_OPEN_SW_HIDDEN,
                       54, 16, x+10, y+30, xgetxt_hidden, &fm->showhidden );
  mw = xge_NewButton ( window_num, mw, xgeid_FILEMENU_OPEN_FILE,
                       58, 19, x+91, y+h-30, xgetxt_Open );
  mw = xge_NewButton ( window_num, mw,  xgeid_FILEMENU_OPEN_CANCEL,
                       58, 19, x+251, y+h-30, xgetxt_Cancel );
  hl = ((h-101)/xge_LISTDIST)*xge_LISTDIST+3;
  mw = xge_NewListBox ( window_num, mw, xgeid_FILEMENU_OPEN_DIRLIST,
                        180, hl, x+10, y+60, &fm->dirlist1 );
  mw = xge_NewListBox ( window_num, mw, xgeid_FILEMENU_OPEN_FILELIST,
                        180, hl, x+210, y+60, &fm->filelist1 );
  fm->er1 = menu = xge_NewFMenu ( window_num, NULL, id, w, h, x, y, mw );
  menu->data2 = fm;
  menu->libmsgproc = _xge_OpenFileProc;
  menu->msgproc = xge_PopupMenuMsg;
  return fm->er1;
} /*xge_InitOpenFileMenu*/

