
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

static const char sh_version400[] =
  "#version 400\n";

static const char sh_gc[] =
  "uniform xgleGC {"
    "mat4 pm;"      /* projection matrix */
    "vec4 fg, bk;"  /* foreground and background */
    "int  h;"       /* viewport height */
  "} xglegc;"
  ;

static const char sh_vert[] =
  "layout(location=0) in vec4 in_Position;"

  "void main ( void )"
  "{"
    "gl_Position = xglegc.pm * in_Position;"
  "}" /*main*/
  ;

static const char sh_line_frag[] =
  "out vec4 out_Color;"

  "void main ( void )"
  "{"
    "out_Color = xglegc.fg;"
  "}" /*main*/
  ;

static const char sh_font_frag_0[] =
  "uniform xgleFont {"
    "int          chw, chh;" /* character width and height */
    "int          chf, chl;" /* first and last character */
    "unsigned int glyphs[3*96];"
  "} xglefont;"

  "uniform xgleText {"
    "int          x, y;"       /* position of the first character */
    "int          l;"          /* string length */
    "unsigned int text[256];"  /* packed characters */
  "} xgletext;"

  "out vec4 out_Color;"

  "void main ( void )"
  "{"
    "int          x0, y0;"
    "unsigned int c, r, mask, chrow;"

    "x0 = int(gl_FragCoord.x) - xgletext.x;"
    "if ( x0 < 0 || x0 >= xglefont.chw*xgletext.l ) {"
      "if ( xglegc.bk.a > 0.0 ) {"
        "out_Color = xglegc.bk;"
        "return;"
      "}"
      "else "
        "discard;"
    "}"
    ;
static const char sh_font_frag_1[] =
    "y0 = xglegc.h - 2 - int(gl_FragCoord.y) - xgletext.y;"
    "if ( y0 < 0 || y0 >= xglefont.chh ) {"
      "if ( xglegc.bk.a > 0.0 ) {"
        "out_Color = xglegc.bk;"
        "return;"
      "}"
      "else "
        "discard;"
    "}"
          /* unpack the character - 4 in one integer */
    "c = xgletext.text[x0/(xglefont.chw*4)];"
    "switch ( (x0/xglefont.chw) % 4 ) {"
  "case 0: break;"
  "case 1: c >>= 8;   break;"
  "case 2: c >>= 16;  break;"
  "case 3: c >>= 24;  break;"
    "}"
    "c &= 0xFF;"
    "if ( c < xglefont.chf || c > xglefont.chl )"
      "c = 0x20;"  /* blank space instead of absent characters */
    "c -= xglefont.chf;"
          /* unpack the glyph line */
    "r = c*xglefont.chh + y0;"
    ;
static const char sh_font_frag_2[] =
    "if ( xglefont.chw <= 8 ) {"
      "chrow = xglefont.glyphs[r/4];"
      "switch ( r % 4 ) {"
    "case 0: break;"
    "case 1: chrow >>= 8;   break;"
    "case 2: chrow >>= 16;  break;"
    "case 3: chrow >>= 24;  break;"
      "}"
    "}"
    "else if ( xglefont.chw <= 16 ) {"
      "chrow = xglefont.glyphs[r/2];"
      "if ( r % 2 != 0 )"
        "chrow >>= 16;"
    "}"
    "else {" /* must be xglefont.chw <= 32 */
      "chrow = xglefont.glyphs[r];"
    "}"
    ;
static const char sh_font_frag_3[] =
          /* draw or discard the pixel */
    "x0 %= xglefont.chw;"
    "mask = 0x01 << x0;"
    "if ( (chrow & mask) != 0 ) {"
      "if ( xglegc.fg.a > 0.0 ) {"
        "out_Color = xglegc.fg;"
        "return;"
      "}"
      "else "
        "discard;"
    "}"
    "else {"
      "if ( xglegc.bk.a > 0.0 ) {"
        "out_Color = xglegc.bk;"
        "return;"
      "}"
      "else "
        "discard;"
    "}"
  "}" /*main*/
  ;

static const char *_shader_v[] =
  { sh_version400, sh_gc, sh_vert };

static const char *_shader_f0[] =
  {sh_version400, sh_gc, sh_line_frag };

static const char *_shader_f1[] =
  {sh_version400, sh_gc, sh_font_frag_0, sh_font_frag_1,
   sh_font_frag_2, sh_font_frag_3 };


