
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "xgledit.h"
#include "xgeprivate.h"

GLuint xgle_CompileShaderStrings ( GLenum shader_type, int nsl,
                                   const GLchar *const*srclines )
{
  void   *sp;
  GLuint shader_id;
  GLint  logsize;
  GLchar *log;

  sp = pkv_GetScratchMemTop ();
        /* compile the entire shader */
  if ( (shader_id = glCreateShader ( shader_type )) != 0 ) {
    glShaderSource ( shader_id, nsl, (const GLchar*const*)srclines, NULL );
    glCompileShader ( shader_id );
    glGetShaderiv ( shader_id, GL_INFO_LOG_LENGTH, &logsize );
    if ( logsize > 1 ) {
      if ( (log = pkv_GetScratchMem ( logsize+1 )) != 0 ) {
        glGetShaderInfoLog ( shader_id, logsize, &logsize, log );
        fprintf ( stderr, "%s\n", log );
      }
    }
  }
  xgle_ExitOnGLError ( "Shader compilation failed" );
  pkv_SetScratchMemTop ( sp );
  return shader_id;
} /*xgle_CompileShaderStrings*/

GLuint xgle_LinkProgram ( int nsh, const GLuint *shaders )
{
  void   *sp;
  GLuint program_id;
  int    i;
  GLint  logsize;
  GLchar *log;

  sp = pkv_GetScratchMemTop ();
  if ( (program_id = glCreateProgram ()) ) {
    for ( i = 0; i < nsh; i++ )
      if ( shaders[i] )
        glAttachShader ( program_id, shaders[i] );
      else
        goto way_out;
    glLinkProgram ( program_id );
    glGetProgramiv ( program_id, GL_INFO_LOG_LENGTH, &logsize );
    if ( logsize > 1 ) {
      if ( (log = pkv_GetScratchMem ( logsize+1 )) ) {
        glGetProgramInfoLog ( program_id, logsize, &logsize, log );
        fprintf ( stderr, "%s\n", log );
      }
    }
  }

way_out:
  xgle_ExitOnGLError ( "Program linking failed" );
  pkv_SetScratchMemTop ( sp );
  return program_id;
} /*xgle_LinkProgram*/

